/*
 * This file is a product of Exabyte Corporation and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify this file without charge, but are not authorized to
 * license or distribute it to anyone else except as part of a product
 * or program developed by the user.
 *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 * This file is provided with no support and without any obligation on the
 * part of Exabyte Corp. to assist in its use, correction, modification or
 * enhancement.
 *
 * EXABYTE CORP. SHALL HAVE NO LIABILITY WITH RESPECT TO THE INFRINGEMENT
 * OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS FILE OR ANY PART
 * THEREOF.
 *
 * In no event will Exabyte Corporation be liable for any lost revenue or
 * profits or other special, indirect and consequential damages, even if
 * Exabyte has been advised of the possibility of such damages.
 *
 * Exabyte Corporation
 * 1685 38th Street
 * Boulder, Colorado  80301
 */

#ifndef	_SYS_CHGRIO_H
#define	_SYS_CHGRIO_H

#ifdef	__cplusplus
extern "C" {
#endif


/*
 * Driver info structure.  Returned by CHGRIOC_DRIVERINFO ioctl.
 */
typedef struct chgr_driver_info {
	char	di_name[32];	/* driver name */
	u_short	di_vers_major;	/* major version num */
	u_short	di_vers_minor;	/* minor version num */
	char	di_vendor[32];	/* driver vendor */
} CHGR_DRVRINFO, *pCHGR_DRVRINFO;


/*
 * Driver features structure.  Returned by CHGRIOC_GET_FEATURES ioctl.
 */
typedef struct chgr_driver_features {
	u_char	df_get_info;          /* CHGRIOC_DRIVERINFO supported */
	u_char	df_get_features;      /* CHGRIOC_GET_FEATURES supported */
	u_char	df_get_capabilities;  /* CHGRIOC_DEVICE_CAPABILITIES */
	u_char	df_get_addresses;     /* CHGRIOC_ELEMENT_ADDRESSES supported */
	u_char	df_set_addresses;     /* CHGRIOC_SET_ELEM_ADDR supported */
	u_char	df_inquiry;	      /* CHGRIOC_INQUIRY supported */
	u_char	df_init;	      /* CHGRIOC_INIT_ELEMENT supported */
	u_char	df_elem_status;	      /* CHGRIOC_ELEM_STATUS supported */
	u_char	df_reserve;	      /* CHGRIOC_RESERVE supported */
	u_char	df_release;	      /* CHGRIOC_RELEASE supported */
	u_char	df_lock;	      /* CHGRIOC_LOCK supported */
	u_char	df_move;	      /* CHGRIOC_MOVE supported */
	u_char	df_position;	      /* CHGRIOC_POSITION supported */
	u_char	df_exchange;	      /* CHGRIOC_EXCHANGE supported */
	u_char	df_send_vtag;	      /* CHGRIOC_SENS_VOL_TAG supported */
	u_char	df_req_voladdr;	      /* CHGRIOC_REQ_VOL_ADDR supported */
	u_char	df_reset;	      /* CHGRIOC_RESETDEVICE supported */
	u_char	df_get_sense;	      /* CHGRIOC_GETSENSE supported */
	u_char	df_pass_through;      /* CHGRIOC_RAWSCSI supported */
	u_char	df_flag;	      /* driver flags */
} CHGR_FEATURES, *pCHGR_FEATURES;

/* flags for df_flags field */
#define CHGR_OPEN_EXCL	0x01		/* driver suports exclusive open */


/*
 * Device capabilities structure. 
 */
typedef struct chgr_capabilities {
	u_char	cap_stor[4];	/* array of storage element types    */
	u_char	cap_move[4][4];	/* array of movement element types   */
	u_char	cap_exch[4][4];	/* array of exchange element types   */
	u_char	cap_rotate;	/* rotate media capability flag      */
} CHGR_CAPABILITIES, *pCHGR_CAPABILITIES;

/* defines used for indexing the stor, move & exch arrays.  */
#define	CHGR_STOR_DT		0
#define	CHGR_STOR_IE		1
#define	CHGR_STOR_MT		2
#define	CHGR_STOR_ST		3


/*
 * Medium changer element addresses structure.  Used to report/set the
 * medium changer hardware element addresses.
 */
typedef struct chgr_addresses {
	u_short	adr_mt_address;	/* 1st medium transport elem addr    */
	u_short	adr_mt_number;	/* number of medium transport elems  */
	u_short	adr_st_address;	/* 1st storage element address       */
	u_short	adr_st_number;	/* number of storage elements        */
	u_short	adr_ie_address;	/* 1st import/export element address */
	u_short	adr_ie_number;	/* number of import/export elements  */
	u_short	adr_dt_address;	/* 1st data transfer element address */
	u_short	adr_dt_number;	/* number of data transfer elements  */
	u_short	adr_reserved;
} CHGR_ADDRS, *pCHGR_ADDRS;


/*
 * Set element addresses request parameters
 */
typedef struct chgr_setaddr_params {
	u_char	sa_restore;	/* restore default addresses */
	u_char	sa_save;	/* store new addr assignments */
	char	*sa_addr_ptr;	/* addr assignment struct ptr */
	u_short	sa_addr_size;	/* size of new addresses struct */
	u_char	sa_control;	/* vendor unique control data */
} CHGR_SETADDR_PARAMS, *pCHGR_SETADDR_PARAMS;


/*
 * Inquiry request parameters
 */
typedef struct chgr_inquiry_params {
	u_char	inqp_evpd;	/* enable vital product data */
	u_char	inqp_pcode;	/* vital product data page code */
	char	*inqp_inq_ptr;	/* ptr to buf to recv inq data */
	u_short	inqp_inq_size;	/* size of inq data buffer */
	u_char	inqp_control;	/* vendor unique control data */
} CHGR_INQ_PARAMS, *pCHGR_INQ_PARAMS;

#define VIDSZ	8
#define PIDSZ	16
#define REVSZ	4
/*
 * SCSI Inquiry Data
 *
 * Format of data returned as a result of an INQUIRY command.
 * The actual values vary contingent on the contents of the
 * inq_rdf field- RDF_LEVEL0 means that only the first 4 bytes
 * are valid (but perhaps not even ISO, ECMA, or ANSI- essentially
 * all you can trust with level 0 is the the device type and whether
 * or not it is removable media).
 *
 * RDF_CCS means that this structure complies with CCS pseudo-spec.
 *
 * RDF_SCSI2 means that the structure complies with the SCSI-2 spec.
 *
 */
typedef struct chgr_inquiry {
#if defined(_BIT_FIELDS_LTOH)
	u_char	inq_dtype	:5,	/* peripheral device type */
		inq_pqual	:3;	/* peripheral qualifier */
	u_char	inq_dtq		:7,	/* device type qualifier */
		inq_rmb		:1;	/* removable media */
	u_char	inq_ansi	:3,	/* ANSI version */
		inq_ecma	:3,	/* ECMA version */
		inq_iso		:2;	/* ISO version */
	u_char	inq_rdf		:4,	/* response data format */
		inq_resv1	:2,	/* reserved */
		inq_trmiop	:1,	/* TERMINATE I/O PROC msg */
		inq_aenc	:1;	/* async event notification */
	u_char	inq_addlen;		/* additional length */
	u_char	inq_resv2[2];		/* reserved */
	u_char	inq_sftre	:1,	/* supports Soft Reset */
		inq_cmdque	:1,	/* supports command queueing */
		inq_resv3	:1,	/* reserved */
		inq_linked	:1,	/* supports linked commands */
		inq_sync	:1,	/* supports synchronous xfers */
		inq_wbus16	:1,	/* supports 16 bit wide xfers */
		inq_wbus32	:1,	/* supports 32 bit wide xfers */
		inq_reladdr	:1;	/* supports relative addr */
	char	inq_vid[VIDSZ];		/* vendor ID */
	char	inq_pid[PIDSZ];		/* product ID */
	char	inq_rev[REVSZ];		/* revision level */
	/*
	 * Bytes 36-55 are vendor-specific.
	 */
	u_char	inq_vuniq[20];		/* device specific */
	/*
	 * Bytes 56-95 are reserved.
	 * 96 to 'n' are vendor-specific parameter bytes
	 */
#elif defined(_BIT_FIELDS_HTOL)
	u_char	inq_pqual	:3,	/* peripheral qualifier */
		inq_dtype	:5;	/* peripheral device type */
	u_char	inq_rmb		:1,	/* removable media */
		inq_dtq		:7;	/* device type qualifier */
	u_char	inq_iso		:2,	/* ISO version */
		inq_ecma	:3,	/* ECMA version */
		inq_ansi	:3;	/* ANSI version */
	u_char	inq_aenc	:1,	/* async event notification */
		inq_trmiop	:1,	/* supports term I/O proc */
		inq_resv1	:2,	/* reserved */
		inq_rdf		:4;	/* response data format */
	u_char	inq_addlen;		/* additional length */
	u_char	inq_resv2[2];		/* reserved */
	u_char	inq_reladdr	:1,	/* supports relative addr */
		inq_wbus32	:1,	/* supports 32 bit wide xfers */
		inq_wbus16	:1,	/* supports 16 bit wide xfers */
		inq_sync	:1,	/* supports synchronous xfers */
		inq_linked	:1,	/* supports linked commands */
		inq_resv3	:1,	/* reserved */
		inq_cmdque	:1,	/* supports command queueing */
		inq_sftre	:1;	/* supports Soft Reset option */
	char	inq_vid[VIDSZ];		/* vendor ID */
	char	inq_pid[PIDSZ];		/* product ID */
	char	inq_rev[REVSZ];		/* revision level */
	/*
	 * Bytes 36-55 are vendor-specific.
	 */
	u_char	inq_vuniq[20];		/* device specific */
	/*
	 * Bytes 56-95 are reserved.
	 * 96 to 'n' are vendor-specific parameter bytes
	 */
#else
#error	One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif	/* _BIT_FIELDS_LTOH */
} CHGR_INQUIRY, *pCHGR_INQUIRY;

/*
 * Initialize element status Request parameters
 */
typedef struct chgr_init_params {
	u_char	ini_control;	/* vendor unique control data */
} CHGR_INIT_PARAMS, *pCHGR_INIT_PARAMS;

/*
 * Input parameter structure for CHGRIOC_ELEM_STATUS and
 * CHGRIOC_REQ_VOL_ADDR ioctls
 */
typedef struct chgr_element_status_params {
	u_char	es_type;	/* element type code */
	u_short	es_addr;	/* starting element address */
	u_short	es_nelems;	/* num elements descs to return */
	u_char	es_tag;		/* include voltag info */
	char	*es_status_p;	/* buffer to recv status data */
	u_long	es_status_size;	/* num bytes alloced for status data */
	u_char	es_control;	/* vendor unique control data */
} CHGR_ESTATUS_PARAMS, *pCHGR_ESTATUS_PARAMS;

/*
 * defines for element status page type codes
 */
#define CHGR_EST_ALL	0x00	/* all status pages */
#define CHGR_EST_MT	0x01	/* transport element status pages */
#define CHGR_EST_ST	0x02	/* sotrage element status pages */
#define CHGR_EST_IE	0x03	/* in/export element status pages */
#define CHGR_EST_DT	0x04	/* data transport element status pages */

/*
 * Structures for reporting element status information, returned by the
 * CHGRIOC_ELEMENT_STATUS and CHGRIOC_REQ_VOL_ADDR ioctls.
 *
 * status info consists of a status header, a status page (one for each
 * element type returned, followed an element status descriptor (one for
 * each element reported). 
 */

/*
 * Element Status Header
 */
typedef struct chgr_element_status_hdr {
	u_short	esh_element;		/* starting elem address */
	u_short	esh_nelems;		/* num elements */
#if defined(_BIT_FIELDS_LTOH)
	u_char	esh_acode	:5,	/* action code */
		esh_reserved	:3;
#elif defined(_BIT_FIELDS_HTOL)
	u_char	esh_reserved	:3,
		esh_acode	:5;	/* action code */
#else
#error	One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif	/* _BIT_FIELDS_LTOH */
	u_char	esh_nbytes[3];		/* num of status bytes */
} CHGR_ESTATUS_HDR, *pCHGR_ESTATUS_HDR;
#define	STATUS_HDR_LENGTH	(sizeof(CHGR_ESTATUS_HDR))

/*
 * element status page
 */
typedef struct chgr_element_status_page {
	u_char	esp_type;		/* element type code */
#if defined(_BIT_FIELDS_LTOH)
	u_char	esp_resv1	:6,
		esp_avoltag	:1,	/* alt volume tag present */
		esp_pvoltag	:1;	/* pri volume tag present */
#elif defined(_BIT_FIELDS_HTOL)
	u_char	esp_pvoltag	:1,	/* pri volume tag present */
		esp_avoltag	:1,	/* alt volume tag present */
		esp_resv1	:6;
#else
#error	One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif	/* _BIT_FIELDS_LTOH */
	u_short	esp_length;		/* num bytes in a elem desc */
	u_char	esp_reserved;
	u_char	esp_nbytes[3];		/* num bytes desc data */
} CHGR_ESTATUS_PAGE, *pCHGR_ESTATUS_PAGE;
#define	STATUS_PAGE_LENGTH	(sizeof(CHGR_ESTATUS_PAGE))

/*
 * element status descriptor
 */
typedef struct chgr_element_status_desc {
#if defined(_BIT_FIELDS_LTOH)
	u_short	esd_element;		/* element address */
	u_char	esd_full	:1,	/* element contains media */
		esd_operatr	:1,	/* operator positioned media */
		esd_except	:1,	/* is in an abnormal state */
		esd_access	:1,	/* MT element accessible */
		esd_ex_enab	:1,	/* allows media removal */
		esd_in_enab	:1,	/* allows media insertion */
		esd_resv1	:2;
	u_char	esd_resv2;
	u_char	esd_asc;		/* info about abnormal state */
	u_char	esd_asq;		/* info about abnormal state */
	u_char	esd_lun		:3,	/* DT device SCSI LUN */
		esd_resv4	:1,
		esd_lu_valid	:1,	/* lun is valid */
		esd_id_valid	:1,	/* bus address is valid */
		esd_resv3	:1,
		esd_not_bus	:1;	/* DT device SCSI bus differs */
	u_char	esd_bus_addr;		/* DT device SCSI address */
	u_char	esd_resv5;
	u_char	esd_resv6	:6,
		esd_invert	:1,	/* medium was inverted */
		esd_svalid	:1;	/* src_addr is valid */
	u_short	esd_src_addr;		/* last storage location */
	char	esd_pri_voltag[36];	/* primary volume tag */
	char	esd_alt_voltag[36];	/* alternate volume tag */
	u_char	esd_misc_bytes[168];	/* device specific */
#elif defined(_BIT_FIELDS_HTOL)
	u_short	esd_element;		/* element address */
	u_char	esd_resv1	:2,
		esd_in_enab	:1,	/* allows media insertion */
		esd_ex_enab	:1,	/* allows media removal */
		esd_access	:1,	/* MT element accessible */
		esd_except	:1,	/* is in an abnormal state */
		esd_operatr	:1,	/* operator positioned media */
		esd_full	:1;	/* element contains media */
	u_char	esd_resv2;
	u_char	esd_asc;		/* info about abnormal state */
	u_char	esd_asq;		/* info about abnormal state */
	u_char	esd_not_bus	:1,	/* DT device SCSI bus differs */
		esd_resv3	:1,
		esd_id_valid	:1,	/* bus address is valid */
		esd_lu_valid	:1,	/* lun is valid */
		esd_resv4	:1,
		esd_lun		:3;	/* DT device SCSI LUN */
	u_char	esd_bus_addr;		/* DT device SCSI address */
	u_char	esd_resv5;
	u_char	esd_svalid	:1,	/* src_addr is valid */
		esd_invert	:1,	/* medium was inverted */
		esd_resv6	:6;
	u_short	esd_src_addr;		/* last storage location */
	char	esd_pri_voltag[36];	/* primary volume tag */
	char	esd_alt_voltag[36];	/* alternate volume tag */
	u_char	esd_misc_bytes[168];	/* device specific */
#else
#error	One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif	/* _BIT_FIELDS_LTOH */
} CHGR_ESTATUS_DESC, *pCHGR_ESTATUS_DESC;
#define	STATUS_DECHGR_LENGTH	(sizeof(CHGR_ESTATUS_DESC))


/*
 * element list descriptor, for specifing a block of elements to
 * be reserved.
 */
typedef struct chgr_elist_desc {
	u_short	elist_reserved;		/* reserved */
	u_short	elist_nelems;		/* num of elements */
	u_short	elist_start_addr;	/* starting elem addr */
} CHGR_ELIST_DESC, *pCHGR_ELIST_DESC;
#define ELIST_DECHGR_LENGTH	(sizeof(CHGR_ELIST_DESC))


/*
 * Reserve/Release request parameters
 *
 * Note:	set elem_resv field to zero (0) to reserve/release entire unit
 * 		(i.e., all elements).
 */
typedef struct chgr_reservation_params {
	u_char	resv_type;		/* reserve unit or elem list */
	u_char	resv_id;		/* reservation ID */
	u_char	resv_third_pty;		/* 3rd party reservation flag */
	u_char	resv_third_pty_id;	/* third party device id */
	char	*resv_elist_ptr;	/* ptr to element list */
	u_short	resv_elist_size;	/* element list size */
	u_char	resv_control;		/* vendor unique control data */
} CHGR_RESV_PARAMS, *pCHGR_RESV_PARAMS;

/* defines for resv_type field */
#define CHGR_RESV_UNIT	0	/* Reserve/release entire unit */
#define CHGR_RESV_LIST	1	/* Reserve/release list of elements */


/*
 * Prevent/Allow media removal request parameters
 */
typedef struct chgr_lock_params {
	u_char	lock_lcode;	/* lock code */
	u_char	lock_control;	/* vendor unique control data */
} CHGR_LOCK_PARAMS, *pCHGR_LOCK_PARAMS;

/* defines for lock_lcode field */
#define CHGR_LC_UNLOCK	0x00	/* allow medium removal */
#define CHGR_LC_LOCK	0x01	/* prevent medium removal */


/*
 * Move media request parameters
 */
typedef struct chgr_move_medium_params {
	u_short	mv_transport;	/* transport element address */
	u_short	mv_source;	/* source element address */
	u_short	mv_dest;	/* destination element address */
	u_char	mv_invert;	/* invert medium on move flag */
	u_char	mv_control;	/* vendor unique control data */
} CHGR_MOVE_PARAMS, *pCHGR_MOVE_PARAMS;


/*
 * Position to element request parameters
 */
typedef struct chgr_position_params {
	u_short	pos_transport;	/* transport element address */
	u_short	pos_dest;	/* destination element address */
	u_char	pos_invert;	/* invert flag */
	u_char	pos_control;	/* vendor unique control data */
} CHGR_POS_PARAMS, *pCHGR_POS_PARAMS;


/*
 * Exchange media request parameters
 */
typedef struct chgr_exchange_medium_params {
	u_short	exch_transport;		/* transport element address */
	u_short	exch_source;		/* source element address */
	u_short	exch_first_dest;	/* first destination element addr */
	u_short	exch_second_dest;	/* second destination element addr */
	u_char	exch_inv1;		/* invert first medium flag */
	u_char	exch_inv2;		/* invert second medium flag */
	u_char	exch_control;		/* vendor unique control data */
} CHGR_EXCH_PARAMS, *pCHGR_EXCH_PARAMS;


/*
 * Send volume tag template structure
 */
typedef struct	chgr_tag_info {
	char	ti_name[32];	/* vol id template */
	u_short	ti_resv1;
	u_short	ti_min_seq;	/* min sequence num */
	u_short ti_resv2;
	u_short	ti_max_seq;	/* max sequence num */
} CHGR_TAG_INFO, *pCHGR_TAG_INFO;

/*
 * Send volume tag request parameters structure
 */
typedef struct chgr_send_vtag_params {
	u_char		svt_etype;	/* element type code */
	u_short		svt_eaddr;	/* element address */
	u_char		svt_acode;	/* send action code */
	CHGR_TAG_INFO	svt_info;	/* volume tag info */
	u_char		svt_control;	/* vendor specific */
} CHGR_SVTAG_PARAMS, *pCHGR_SVTAG_PARAMS;

/*
 * Send action codes
 */
#define CHGR_ACODE_TRANS_SEARCH_ALL		0x00
#define CHGR_ACODE_TRANS_SEARCH_PRI		0x01
#define CHGR_ACODE_TRANS_SEARCH_ALT		0x02
/* 0x03 reserved */
#define CHGR_ACODE_TRANS_SEARCH_ALL_NOSEQ	0x04
#define CHGR_ACODE_TRANS_SEARCH_PRI_NOSEQ	0x05
#define CHGR_ACODE_TRANS_SEARCH_ALT_NOSEQ	0x06
/* 0x07 reserved */
#define CHGR_ACODE_ASSERT_PRI			0x08
#define CHGR_ACODE_ASSERT_ALT			0x09
#define CHGR_ACODE_REPLACE_PRI			0x0A
#define CHGR_ACODE_REPLACE_ALT			0x0B
#define CHGR_ACODE_UNDEF_PRI			0x0C
#define CHGR_ACODE_UNDEF_ALT			0x0D
/*
 * 0x0E through 0x1B reserved.
 * 0x1C through 0x1F vendor specific.
 */

/*
 * Standard (Non-Extended) SCSI Sense.
 *
 * For Error Classe 0-6. This is all
 * Vendor Unique sense information.
 */
typedef struct chgr_normal_sense {
#if defined(_BIT_FIELDS_LTOH)
	u_char	ns_code		: 4,	/* Vendor Uniqe error code */
		ns_class	: 3,	/* Error class */
		ns_valid	: 1;	/* Logical Blk Addr is valid */
	u_char	ns_lba_hi	: 5,	/* High Logical Blk Addr */
		ns_vu		: 3;	/* Vendor Unique value */
#elif defined(_BIT_FIELDS_HTOL)
	u_char	ns_valid	: 1,	/* Logical Blk Addr is valid */
		ns_class	: 3,	/* Error class */
		ns_code		: 4;	/* Vendor Uniqe error code */
	u_char	ns_vu		: 3,	/* Vendor Unique value */
		ns_lba_hi	: 5;	/* High Logical Blk Addr */
#else
#error	One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif	/* _BIT_FIELDS_LTOH */
	u_char	ns_lba_mid;		/* Middle Logical Blk Addr */
	u_char	ns_lba_lo;		/* Low Logical Blk Addr */
} CHGR_NORMAL_SENSE, *pCHGR_NORMAL_SENSE;

#define	CHGR_CLASS_EXTENDED_SENSE	0x7	/* indicates extended sense */

/*
 * SCSI Extended Sense structure
 *
 * For Error Class 7, the Extended Sense Structure is applicable.
 *
 */
typedef struct chgr_extended_sense {
#if defined(_BIT_FIELDS_LTOH)
	u_char	es_code		: 4,	/* Vendor Unique error code */
		es_class	: 3,	/* Error Class- fixed at 0x7 */
		es_valid	: 1;	/* sense data is valid */
	u_char	es_segnum;		/* segment number */
	u_char	es_key		: 4,	/* Sense key (see below) */
				: 1,	/* reserved */
		es_ili		: 1,	/* Incorrect Length Indicator */
		es_eom		: 1,	/* End of Media */
		es_filmk	: 1;	/* File Mark Detected */
#elif defined(_BIT_FIELDS_HTOL)
	u_char	es_valid	: 1,	/* sense data is valid */
		es_class	: 3,	/* Error Class- fixed at 0x7 */
		es_code		: 4;	/* Vendor Unique error code */
	u_char	es_segnum;		/* segment number */
	u_char	es_filmk	: 1,	/* File Mark Detected */
		es_eom		: 1,	/* End of Media */
		es_ili		: 1,	/* Incorrect Length Indicator */
				: 1,	/* reserved */
		es_key		: 4;	/* Sense key (see below) */
#else
#error	One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif	/* _BIT_FIELDS_LTOH */
	u_char	es_info_1;		/* information byte 1 */
	u_char	es_info_2;		/* information byte 2 */
	u_char	es_info_3;		/* information byte 3 */
	u_char	es_info_4;		/* information byte 4 */
	u_char	es_add_len;		/* num additional bytes */
	u_char	es_cmd_info[4];		/* command specific info */
	u_char	es_add_code;		/* Additional Sense Code */
	u_char	es_qual_code;		/* ASC Qualifier */
	u_char	es_fru_code;		/* Field Replaceable Unit */
	u_char	es_skey_specific[3];	/* Sense Key Specific info */
	/*
	 * Additional bytes may be defined in each implementation.
	 * The actual amount of space allocated for Sense Information
	 * is also implementation dependent.
	 *
	 * Modulo that, the declaration of an array two bytes in size
	 * nicely rounds this entire structure to a size of 20 bytes.
	 */
	u_char	es_add_info[2];		/* additional information */
} CHGR_EXTENDED_SENSE, *pCHGR_EXTENDED_SENSE;

/*
 * Sense Key values for Extended Sense.
 */
#define	CHGR_ES_KEY_NO_SENSE		0x00
#define	CHGR_ES_KEY_RECOVERABLE_ERROR	0x01
#define	CHGR_ES_KEY_NOT_READY		0x02
#define	CHGR_ES_KEY_MEDIUM_ERROR	0x03
#define	CHGR_ES_KEY_HARDWARE_ERROR	0x04
#define	CHGR_ES_KEY_ILLEGAL_REQUEST	0x05
#define	CHGR_ES_KEY_UNIT_ATTENTION	0x06
#define	CHGR_ES_KEY_WRITE_PROTECT	0x07
#define	CHGR_ES_KEY_DATA_PROTECT	ES_KEY_WRITE_PROTECT
#define	CHGR_ES_KEY_BLANK_CHECK		0x08
#define	CHGR_ES_KEY_VENDOR_UNIQUE	0x09
#define	CHGR_ES_KEY_COPY_ABORTED	0x0A
#define	CHGR_ES_KEY_ABORTED_COMMAND	0x0B	/* often a parity error */
#define	CHGR_ES_KEY_EQUAL		0x0C
#define	CHGR_ES_KEY_VOLUME_OVERFLOW	0x0D
#define	CHGR_ES_KEY_MISCOMPARE		0x0E
#define	CHGR_ES_KEY_RESERVED		0x0F


/*
 * user-scsi command structure (i.e., pass-thru commands).
 */
typedef struct chgr_uscsi_cmd {
	int	uscsi_flags;		/* read, write, etc. see below */
	short	uscsi_status;		/* resulting status  */
	short	uscsi_timeout;		/* Command Timeout */
	caddr_t	uscsi_cdb;		/* cdb to send to target */
	caddr_t	uscsi_bufaddr;		/* i/o source/destination */
	u_int	uscsi_buflen;		/* size of i/o to take place */
	u_int	uscsi_resid;		/* resid from i/o operation */
	u_char	uscsi_cdblen;		/* # of valid cdb bytes */
	u_char	uscsi_rqlen;		/* size of uscsi_rqbuf */
	u_char	uscsi_rqstatus;		/* status of request sense cmd */
	u_char	uscsi_rqresid;		/* resid of request sense cmd */
	caddr_t	uscsi_rqbuf;		/* request sense buffer */
	void   *uscsi_reserved_5;	/* Reserved for Future Use */
} CHGR_USCSI_CMD, *pCHGR_USCSI_CMD;

/*
 * flags for uscsi_flags field
 */
#define	CHGR_USCSI_WRITE	0x00000	/* send data to device */
#define	CHGR_USCSI_SILENT	0x00001	/* no error messages */
#define	CHGR_USCSI_DIAGNOSE	0x00002	/* fail if any error occurs */
#define	CHGR_USCSI_ISOLATE	0x00004	/* isolate from normal commands */
#define	CHGR_USCSI_READ		0x00008	/* get data from device */
#define	CHGR_USCSI_NOPARITY	0x00010	/* run command without parity */
#define	CHGR_USCSI_NODISCON	0x00020	/* run command without disconnects */
#define	CHGR_USCSI_NOINTR	0x00040	/* No interrupts */
#define	CHGR_USCSI_NOTAG	0x00100	/* Disable tagged queueing */
#define	CHGR_USCSI_OTAG		0x00200	/* ORDERED QUEUE tagged cmd */
#define	CHGR_USCSI_HTAG		0x00400	/* HEAD OF QUEUE tagged cmd */
#define	CHGR_USCSI_HEAD		0x00800	/* Head of HA que */
#define	CHGR_USCSI_ASYNC	0x01000	/* Set bus to asynchronous mode */
#define	CHGR_USCSI_SYNC		0x02000	/* Return bus to sync mode */
#define	CHGR_USCSI_RESET	0x04000	/* Reset target */
#define	CHGR_USCSI_RESET_ALL	0x08000	/* Reset all targets */
#define	CHGR_USCSI_RQENABLE	0x10000	/* Enable Request Sense extensions */

#define	CHGR_USCSI_RESERVED	0xfffe0000	/* Reserved Bits, must be 0 */


/*
 * Mode sense/select page codes
 */
#define CHGR_PC_PARITY	0x00	/* parity mode page                  */
#define CHGR_PC_ADDR	0x1D	/* element address mode page         */
#define CHGR_PC_GEOM	0x1E	/* transport geometry mode page      */
#define CHGR_PC_DEVCAP	0x1f	/* device capabilities mode page     */
#define CHGR_PC_ALL	0x3F	/* all mode pages                    */

/*
 * Mode sense/select, Parameter list header
 */
typedef struct chgr_mode_param_hdr {
	u_char	mode_data_length;	/* mode data length */
	u_char	mode_reserved[3];
} CHGR_MODE_PARAM_HDR, *pCHGR_MODE_PARAMS_HDR;

#define	MODE_HDR_SIZE	sizeof(CHGR_MODE_PARAM_HDR)

/*
 * Mode sense/select, common page header
 */
typedef struct chgr_mode_page_hdr {
#if defined(_BIT_FIELDS_LTOH)
	u_char	mph_pcode	:6,	/* page code */
		mph_resv	:1,	/* reserved */
		mph_ps		:1;	/* page saveable */
#elif defined(_BIT_FIELDS_HTOL)
	u_char	mph_ps		:1,	/* page saveable */
		mph_resv	:1,	/* reserved */
		mph_pcode	:6;	/* page code */
#else
#error	One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif	/* _BIT_FIELDS_LTOH */
	u_char	mph_page_length;	/* num bytes page data */
} CHGR_MODE_PAGE_HDR, *pCHGR_MODE_PAGE_HDR;

#define PAGE_HDR_SIZE	sizeof(CHGR_MODE_PAGE_HDR)

/*
 * Element Address assignment page.
 *
 * Used to assign addresses to the elements of an independent medium
 * changer (mode select) and to report those assignments (mode sense).
 * This page also defines the number of each type of element present in
 * the medium changer device.
 */
typedef struct chgr_mode_address_page {
	CHGR_MODE_PAGE_HDR	map_page_hdr;	/* common mode page header */
	CHGR_ADDRS		map_page_data;	/* element addresses */
} CHGR_MODE_ADDR_PAGE, *pCHGR_MODE_ADDR_PAGE;

/*
 * Transport geometry parameters page.
 *
 * The transport geometry parameters mode page defines whether each medium
 * transport element of a medium changer is a member of a set of elements
 * that share a common robotics subsystem and whether the element is
 * capable of media rotation.  One transport geometry descriptor is
 * transfered for each medium transport element, begining with the first
 * transport element (other than the default transport element address of
 * zero).
 *
 * The transport geometry parameters page consists of a mode page header
 * and a transport geometry descriptor for each medium transport element
 * in the unit (as reported by the MTNumber field of the element address
 * assignment mode page. 
 */
typedef struct chgr_mode_trans_page {
#if defined(_BIT_FIELDS_LTOH)
	u_char	mtp_rotate	:1,	/* can handle 2-sided media */
		mtp_resv	:7;	/* reserved */
	u_char	mtp_member;
#elif defined(_BIT_FIELDS_HTOL)
	u_char	mtp_resv	:7,	/* reserved */
		mtp_rotate	:1;	/* can handle 2-sided media */
	u_char	mtp_member;
#else
#error	One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif
} CHGR_MODE_TRANS_PAGE, *pCHGR_MODE_TRANS_PAGE;

/*
 * Device capabilities page
 *
 * The device capabilities page defines characteristics of the element
 * types of a medium changer.  This information may be employed by an
 * initiator to determin functions permited by the MOVE MEDIUM and
 * EXCHANGE MEDIUM commands.
 */
typedef struct chgr_mode_devcap_page {
	CHGR_MODE_PAGE_HDR	mcp_page_hdr;	/* common mode page header */
#if defined(_BIT_FIELDS_LTOH)
	u_char	mcp_StorMT	:1,	/* MT elem provides storage */
		mcp_StorST	:1,	/* ST elem provides storage */
		mcp_StorIE	:1,	/* IE elem provides storage */
		mcp_StorDT	:1,	/* DT elem provides storage */
				:4;	/* reserved */
	u_char	mcp_reserved;
	u_char	mcp_MTtoMT	:1,	/* MT to MT move allowed */
		mcp_MTtoST	:1,	/* MT to ST move allowed */
		mcp_MTtoIE	:1,	/* MT to IE move allowed */
		mcp_MTtoDT	:1,	/* MT to DT move allowed */
				:4;	/* reserved */
	u_char	mcp_STtoMT	:1,	/* ST to MT move allowed */
		mcp_STtoST	:1,	/* ST to ST move allowed */
		mcp_STtoIE	:1,	/* ST to IE move allowed */
		mcp_STtoDT	:1,	/* ST to DT move allowed */
				:4;	/* reserved */
	u_char	mcp_IEtoMT	:1,	/* IE to MT move allowed */
		mcp_IEtoST	:1,	/* IE to ST move allowed */
		mcp_IEtoIE	:1,	/* IE to IE move allowed */
		mcp_IEtoDT	:1,	/* IE to DT move allowed */
				:4;	/* reserved */
	u_char	mcp_DTtoMT	:1,	/* DT to MT move allowed */
		mcp_DTtoST	:1,	/* DT to ST move allowed */
		mcp_DTtoIE	:1,	/* DT to IE move allowed */
		mcp_DTtoDT	:1,	/* DT to DT move allowed */
				:4;	/* reserved */
	u_char	mcp_reserved2[4];	/* reserved */
	u_char	mcp_MTxMT	:1,	/* MT to MT exchange allowed */
		mcp_MTxST	:1,	/* MT to ST exchange allowed */
		mcp_MTxIE	:1,	/* MT to IE exchange allowed */
		mcp_MTxDT	:1,	/* MT to DT exchange allowed */
				:4;	/* reserved */
	u_char	mcp_STxMT	:1,	/* ST to MT exchange allowed */
		mcp_STxST	:1,	/* ST to ST exchange allowed */
		mcp_STxIE	:1,	/* ST to IE exchange allowed */
		mcp_STxDT	:1,	/* ST to DT exchange allowed */
				:4;	/* reserved */
	u_char	mcp_IExMT	:1,	/* IE to MT exchange allowed */
		mcp_IExST	:1,	/* IE to ST exchange allowed */
		mcp_IExIE	:1,	/* IE to IE exchange allowed */
		mcp_IExDT	:1,	/* IE to DT exchange allowed */
				:4;	/* reserved */
	u_char	mcp_DTxMT	:1,	/* DT to MT exchange allowed */
		mcp_DTxST	:1,	/* DT to ST exchange allowed */
		mcp_DTxIE	:1,	/* DT to IE exchange allowed */
		mcp_DTxDT	:1,	/* DT to DT exchange allowed */
				:4;	/* reserved */
#elif defined(_BIT_FIELDS_HTOL)
	u_char			:4,	/* reserved */
		mcp_StorDT	:1,	/* DT elem provides storage */
		mcp_StorIE	:1,	/* IE elem provides storage */
		mcp_StorST	:1,	/* ST elem provides storage */
		mcp_StorMT	:1;	/* MT elem provides storage */
	u_char	mcp_reserved;
	u_char			:4,	/* reserved */
		mcp_MTtoDT	:1,	/* MT to DT move allowed */
		mcp_MTtoIE	:1,	/* MT to IE move allowed */
		mcp_MTtoST	:1,	/* MT to ST move allowed */
		mcp_MTtoMT	:1;	/* MT to MT move allowed */
	u_char			:4,	/* reserved */
		mcp_STtoDT	:1,	/* ST to DT move allowed */
		mcp_STtoIE	:1,	/* ST to IE move allowed */
		mcp_STtoST	:1,	/* ST to ST move allowed */
		mcp_STtoMT	:1;	/* ST to MT move allowed */
	u_char			:4,	/* reserved */
		mcp_IEtoDT	:1,	/* IE to DT move allowed */
		mcp_IEtoIE	:1,	/* IE to IE move allowed */
		mcp_IEtoST	:1,	/* IE to ST move allowed */
		mcp_IEtoMT	:1;	/* IE to MT move allowed */
	u_char			:4,	/* reserved */
		mcp_DTtoDT	:1,	/* DT to DT move allowed */
		mcp_DTtoIE	:1,	/* DT to IE move allowed */
		mcp_DTtoST	:1,	/* DT to ST move allowed */
		mcp_DTtoMT	:1;	/* DT to MT move allowed */
	u_char	mcp_reserved2[4];
	u_char			:4,	/* reserved */
		mcp_MTxDT	:1,	/* MT to DT exchange allowed */
		mcp_MTxIE	:1,	/* MT to IE exchange allowed */
		mcp_MTxST	:1,	/* MT to ST exchange allowed */
		mcp_MTxMT	:1;	/* MT to MT exchange allowed */
	u_char			:4,	/* reserved */
		mcp_STxDT	:1,	/* ST to DT exchange allowed */
		mcp_STxIE	:1,	/* ST to IE exchange allowed */
		mcp_STxST	:1,	/* ST to ST exchange allowed */
		mcp_STxMT	:1;	/* ST to MT exchange allowed */
	u_char			:4,	/* reserved */
		mcp_IExDT	:1,	/* IE to DT exchange allowed */
		mcp_IExIE	:1,	/* IE to IE exchange allowed */
		mcp_IExST	:1,	/* IE to ST exchange allowed */
		mcp_IExMT	:1;	/* IE to MT exchange allowed */
	u_char			:4,	/* reserved */
		mcp_DTxDT	:1,	/* DT to DT exchange allowed */
		mcp_DTxIE	:1,	/* DT to IE exchange allowed */
		mcp_DTxST	:1,	/* DT to ST exchange allowed */
		mcp_DTxMT	:1;	/* DT to MT exchange allowed */
#else
#error	One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif
} CHGR_MODE_DEVCAP_PAGE, *pCHGR_MODE_DEVCAP_PAGE;


/*
 * Medium changer supported SCSI commands
 */
#define CHGR_SCMD_UNITREADY		0x00	/* test unit ready           */
#define CHGR_SCMD_REQSENSE		0x03	/* request sense data        */
#define CHGR_SCMD_INITELEMENT		0x07	/* initialize element status */
#define CHGR_SCMD_INQUIRY		0x12	/* inquiry                   */
#define CHGR_SCMD_MODESELECT		0x15	/* mode select               */
#define CHGR_SCMD_RESERVE		0x16	/* researve element/unit     */
#define CHGR_SCMD_RELEASE		0x17    /* release reserved element  */
#define CHGR_SCMD_MODESENSE		0x1A	/* mode sense                */
#define CHGR_SCMD_LOCK			0x1E    /* prevent/allow removal     */
#define CHGR_SCMD_POSITION		0x2B	/* position to element       */
#define CHGR_SCMD_MOVE_MEDIUM		0xA5	/* move medium               */
#define CHGR_SCMD_EXCHANGE_MEDIUM	0xA6	/* exchange medium           */
#define CHGR_SCMD_REQUEST_VOLTAG	0xB5	/* request volume tags       */
#define CHGR_SCMD_SEND_VOLTAG		0xB6	/* send volume tags          */
#define CHGR_SCMD_ELEMSTATUS		0xB8	/* read element status       */

/*
 * Ioctl commands supported by this driver
 */
#define	CHGRIOC		('C' << 8)
#define CHGRIOC_DRIVERINFO		(CHGRIOC | 0x01)
#define CHGRIOC_GET_FEATURES		(CHGRIOC | 0x02)
#define CHGRIOC_DEVICE_CAPABILITIES	(CHGRIOC | 0x03)
#define CHGRIOC_ELEMENT_ADDRESSES	(CHGRIOC | 0x04)
#define CHGRIOC_SET_ELEM_ADDR		(CHGRIOC | 0x05)
#define CHGRIOC_INQUIRY			(CHGRIOC | 0x06)
#define CHGRIOC_INIT_ELEMENT		(CHGRIOC | 0x07)
#define CHGRIOC_ELEM_STATUS		(CHGRIOC | 0x08)
#define CHGRIOC_RESERVE			(CHGRIOC | 0x09)
#define CHGRIOC_RELEASE			(CHGRIOC | 0x0A)
#define CHGRIOC_LOCK			(CHGRIOC | 0x0B)
#define CHGRIOC_MOVE			(CHGRIOC | 0x0C)
#define CHGRIOC_POSITION		(CHGRIOC | 0x0D)
#define CHGRIOC_EXCHANGE		(CHGRIOC | 0x0E)
#define CHGRIOC_SEND_VOLTAG		(CHGRIOC | 0x0F)
#define CHGRIOC_REQ_VOL_ADDR		(CHGRIOC | 0x10)
#define CHGRIOC_RESETDEVICE		(CHGRIOC | 0x11)
#define CHGRIOC_GETSENSE		(CHGRIOC | 0x12)
#define CHGRIOC_RAWSCSI			(CHGRIOC | 0xFF)


#ifdef	__cplusplus
}
#endif

#endif	/* _SYS_CHGRIO_H */
