/*
 * chs.c - Cartridge Handling System 
 *
 * This file is a product of Exabyte Corporation and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify this file without charge, but are not authorized to
 * license or distribute it to anyone else except as part of a product
 * or program developed by the user.
 *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 * This file is provided with no support and without any obligation on the
 * part of Exabyte Corp. to assist in its use, correction, modification or
 * enhancement.
 *
 * EXABYTE CORP. SHALL HAVE NO LIABILITY WITH RESPECT TO THE INFRINGEMENT
 * OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS FILE OR ANY PART
 * THEREOF.
 *
 * In no event will Exabyte Corporation be liable for any lost revenue or
 * profits or other special, indirect and consequential damages, even if
 * Exabyte has been advised of the possibility of such damages.
 *
 * Exabyte Corporation
 * 1685 38th Street
 * Boulder, Colorado  80301
 */

/*
 * chs.c	Command line utility program to manipulate robotic medium
 *		changers.
 *
 * This program uses the Exabyte SCSI Medium Changer device driver.  The
 * driver must be properly installed and configured for this program
 * to function.
 */
#if (defined(solaris) || defined(sparc) || defined(i386))
#if !defined(_SOLARIS)
#define _SOLARIS
#endif

#elif (defined(HPUX10) || defined(HPUX9))
#if !defined(_HPUX)
#define _HPUX
#endif

#endif


#if (defined(_SOLARIS) && (!defined(_HPUX)))
#include <sys/isa_defs.h>	/* SUN platfrom specials */
#endif

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/chgrio.h>

#if (defined(_SOLARIS) && (!defined(_HPUX)))
/* #include <sys/scsi/generic/sense.h> */
#include <sys/scsi/scsi.h>
#endif

/*
 * Local Structures and defines
 */
#define VERS_MAJ	1
#define VERS_MIN	0
#define	BUFSZ		512

#if (defined(_SOLARIS) && (!defined(_HPUX)))
#define DEFAULT_DEV	"/dev/chgr0"
#elif (defined(_HPUX) && (!defined(_SOLARIS)))
#define DEFAULT_DEV	"/dev/rac/c1t6dd"
#endif

char	default_dev[] = DEFAULT_DEV;	/* default medium changer device */

/*
 * Local Macros
 */
#define STVALID(x, y)	(int)(((int)x >= (int)y.adr_st_address) && \
			((int)x < (int)(y.adr_st_address + y.adr_st_number)))
#define DTVALID(x,y)	(int)(((int)x >= (int)y.adr_dt_address) && \
			((int)x < (int)(y.adr_dt_address + y.adr_dt_number)))
#define MTVALID(x,y)	(int)(((int)x >= (int)y.adr_mt_address) && \
			((int)x < (int)(y.adr_mt_address + y.adr_mt_number)))
#define IEVALID(x,y)	(int)(((int)x >= (int)y.adr_ie_address) && \
			((int)x < (int)(y.adr_ie_address + y.adr_ie_number)))
#define ITOS2(i, s)	s[0]= (u_char)(((int)i & 0xff00) >> 8); \
			s[1]= (u_char)((int)i & 0x00ff)
#define S2TOI(s, i)	i = (int) (((int)s[0]) << 8) + (int)s[1]
#define ITOS3(i, s)	s[0]= (u_char)(((int)i & 0xff0000) >> 16); \
			s[1]= (u_char)(((int)i & 0x00ff00) >> 8); \
			s[2]= (u_char)((int)i & 0x0000ff)
#define S3TOI(s, i)	i = (int) (((int)s[0]) << 16) + \
				  (((int)s[1]) <<  8) + \
				    (int)s[2]

/*
 * Function Declarations
 */
void Usage(char *name);
int DoInit();
int DoDriver();
int DoHardware();
int DoReset();
int DoInventory();
int DoSearch();
int DoSelect();
int DoDeselect();
int DoImport();
int DoExport();
int DoPosition();
int DoMove();
int DoExchange();
int DoLock();
int DoUnlock();
void GetSense();
void PrintElementInfo(caddr_t);
void strntrm(char *, int);
int supports_barc();
int GetAddrInfo(CHGR_ADDRS *);
#ifdef DEBUG
int DoRezero();
int DoInquiry();
void hex_print(char *, int);
#endif /*DEBUG*/

/*
 * Global Data
 */
typedef struct CommandTbl {
	char	*Cmd;
	char	*ArgTxt;
	char	*CmdTxt;
	char	*CmdAbbr;
	int	(*fpProcessCmd)();
} CMD_TBL;

CMD_TBL Commands[] = {
	"driver", "", "Display driver information", "dri", DoDriver,
	"hardware", "", "Display changer information", "har", DoHardware,
	"init", "", "Initialize changer unit", "ini", DoInit,
	"inventory","[-l]", "List available media", "inv", DoInventory,
	"search", "-n \"search-str\"", "Search media by name", "sea", DoSearch,
	"select", "-s slot -d drive", "Move media from storage to drive", "sel", DoSelect,
	"deselect", "[-s slot] -d drive", "Move media from drive to storage", "des", DoDeselect,
	"import", "-s slot | -d drive", "Import media to slot or drive", "imp", DoImport,
	"export", "-s slot | -d drive", "Export media from slot or drive", "exp", DoExport,
	"position", "[-s slot | -d drive]", "Position transport element", "pos", DoPosition,
	"move", "-s src -s dest", "Move media from slot to slot", "mov", DoMove,
	"exchange", "-s src -s dest [-s dest]", "Exchange medium locations", "exc", DoExchange,
	"lock", "", "Prevent media access/removal", "loc", DoLock,
	"unlock", "", "Allow media access/removal", "unl", DoUnlock,
	"reset", "", "Reset changer device", "res", DoReset,
#ifdef DEBUG
	"rezero", "", "Rezero unit", "rez", DoRezero,
	"inquiry", "", "Get device inquiry data", "inq", DoInquiry,
#endif /*DEBUG*/
NULL, NULL, NULL, NULL
};

int	Invert = 0;		/* invert media flag */
int	Labels = 0;		/* include vtag info in inventory */
int	Barcode = 0;		/* supports barcode labels */
long	STElement = -1;		/* First storage element address */
long	STElementTwo = -1;	/* Second storage element address */
long	STElementThree = -1;	/* Third storage element address */
long	DTElement = -1;		/* Data transfer element address */
char	Vol_Label[32];		/* volume label template */
char	*Search_Str = NULL;	/* search string ptr */
char	*progname = NULL;	/* the name of this program */
char	buf[BUFSZ];		/* command data buffer */
int	(*fpCommand)();		/* command to execute */
int	fd;			/* file descriptor handle */
CHGR_ADDRS	Addresses;	/* element addresses */


int main(int argc, char **argv)
{
	extern char	*optarg;
	extern int	optind;
	char		*devname = NULL;
	char		*cpCmd,
			*cpCmdAbbr;
	long		st;
	long		dt;	
	int		c, i, err;


	progname = argv[0];	/* name of this program */
	if (argc < 2) {
		(void) Usage(argv[0]);
		return (1);
	}

	/*
	 * Check for device handle
	 */
	if ((c=getopt(argc, argv, "f:")) == 'f') {
		/*
		 * A device name has been specified on commandline,
		 * use it instead of the default.
		 */
		devname= optarg;
	} else {
		/*
		 * check for CHANGER environment variable.  If
		 * present and set, use it.
		 */
		devname= getenv("CHANGER");
	}

	/*
	 * Still have'nt found a changer device?  Use the default.
	 */
	if (devname == NULL) 
		devname = default_dev;

	/*
	 * Get the command from the argument list
	 */
	if ((optind >= argc) || (NULL == argv[optind])) {
		/* Nothing left on the command line for a command */
		(void) Usage(argv[0]);
		return (1);
	}
	for (i= 0, fpCommand= NULL; (Commands[i].Cmd != NULL) &&
				    (fpCommand == NULL); i++) {
		cpCmd = Commands[i].Cmd;
		cpCmdAbbr = Commands[i].CmdAbbr;

		/*
		 * Check for a match of the cmd name or abbreviation
		 */
		if ((!(strncmp(cpCmd,  argv[optind], strlen(cpCmd)))) ||
		    (!(strncmp(cpCmdAbbr,  argv[optind], 3)))) {
			fpCommand = Commands[i].fpProcessCmd;
		}
	}
	if(NULL == fpCommand) {
		(void) fprintf(stderr, "Unknown command\n");
		Usage(argv[0]);  /* No command on command line */
		return (1);
	}


	/*
	 * We now have a valid device handle, and a valid command.
	 * Open the device and probe for setup info.
	 */

	/*
	 * Open the device
	 */
	if ((fd = open(devname, O_RDWR|O_EXCL)) == -1) {
		perror(progname);
		return(1);
	}

	/*
	 * Check for barcode label support (on Exabyte devices)
	 */
	if (supports_barc())
		Barcode = 1;

	/*
	 * get element addresses
	 */
	(void) memset((void *) &Addresses, 0, sizeof(CHGR_ADDRS));
	if (GetAddrInfo(&Addresses)) {
		perror(progname);
		GetSense();
		return(-1);
	}


	/*
	 * Parse the rest of the cmd line for flags
	 */
	optind++;
	while((c=getopt(argc, argv, "d:iln:s:")) != EOF) {
		switch(c) {
		case 'd':	/* DT element address(s) */
			if (DTElement == -1) {
				dt = atol(optarg);
				DTElement = Addresses.adr_dt_address + dt - 1;
			} else {
				(void) Usage(argv[0]);
				return (1);
			}
			break;
		case 'i':	/* invert media on move or exchange	*/
			/*
			 * Note:  Need to perform a check to verify that
			 * media rotation is supported for the target
			 * device.  If not, display warning and proceed
			 * without rotation.
			 */
			Invert++;
			break;
		case 'l':	/* include Vtag info with inventory     */
			/*
			 * Note:  Should include a check to make sure
			 * that the device supports media labels.  This
			 * is currently a vendor uniq field of the inquiry
			 * data, so there is not a sure fire method for
			 * determining this that will work for all medium
			 * changers.
			 */
			Labels++;
			break;
		case 'n':	/* named search string */
			strncpy(Vol_Label, optarg, 31);
			Vol_Label[31] = 0;
			Search_Str = Vol_Label;
			break;
		case 's':	/* source element address */
			if (STElement == -1) {
				st = atol(optarg);
				STElement = Addresses.adr_st_address + st - 1;
			} else if (STElementTwo == -1) {
				/*
				 * can only have 2 ST elements if doing
				 * a move or exchange.
				 */
				st = atol(optarg);
				STElementTwo= Addresses.adr_st_address+st - 1;
			} else if (STElementThree == -1) {
				/*
				 * can only have 3 ST elements if doing
				 * an exchange.
				 */
				st = atol(optarg);
				STElementThree= Addresses.adr_st_address+st - 1;
			} else {
				(void) Usage(argv[0]);
				return (1);
			}
			break;
		case '?':
		default:
			(void) Usage(argv[0]);
			return (1);
		}
	}


	/*
	 * execute the command, then close the device and exit.
	 */
	err= (fpCommand)();
	(void) close(fd);

	return(err);
}

/*
 * Display usage message
 */
void Usage(char *name)
{
	int	i;

	(void) fprintf(stderr, "%s version %d.%d\n", name, VERS_MAJ, VERS_MIN);
	(void) fprintf(stderr, "usage: %s [-f changername] command [-s slot] "
		"[-d drive] [-n label] [-l]\n", name);
	(void) fprintf(stderr, "Commands:\n");
	for (i=0; Commands[i].Cmd != NULL; i++) {
		(void) fprintf(stderr, "\t%-9s %-24s - %s\n", Commands[i].Cmd,
				Commands[i].ArgTxt, Commands[i].CmdTxt);
	}
}

#define NBL	0x80
/*
 * Initialize element status for all elements
 */
int DoInit()
{
	CHGR_INIT_PARAMS	InitReq;

	(void) memset((void *) &InitReq, 0, sizeof(CHGR_INIT_PARAMS));

	/*
	 * If the changer supports a barcode reader, then by
	 * default, we will scan the labels during init.
	 *
	 * If barcode labels supported, and we do NOT want to
	 * scan for barcodes during the init, then we must specify
	 * the NBL control bit. If barcode labels are not supported
	 * and the NBL bit is set, then we get an error.
	 *
	 * On the other hand, if we do not set the NBL bit, the
	 * command works fine regardless of whether or not the
	 * device supports barcode labels.  Scanning the barcodes
	 * on devices which do this can take a long time however.
	 * 
	 * InitReq.ini_control = (Barcode && Labels)?NBL:0;
	 */
	if (ioctl(fd, CHGRIOC_INIT_ELEMENT, &InitReq) == -1) {
		perror(progname);
		GetSense();
		return(2);
	}
	return (0);
}

/*
 * Get Driver information and display
 */
int DoDriver()
{
	CHGR_DRVRINFO	*DrvrInfo;

	(void) memset((void *) buf, 0, sizeof(CHGR_DRVRINFO));
	if (ioctl(fd, CHGRIOC_DRIVERINFO, (void *)buf) == -1) {
		perror(progname);
		return(2);
	}
	DrvrInfo = (CHGR_DRVRINFO *) buf;
	(void) fprintf(stdout, "%s Version %d.%d by %s\n",
		DrvrInfo->di_name, DrvrInfo->di_vers_major,
		DrvrInfo->di_vers_minor, DrvrInfo->di_vendor);
	return (0);
}

/*
 * Get hardware information & capabilities
 */
int DoHardware()
{
	CHGR_INQ_PARAMS	InqReq;
	CHGR_INQUIRY	Inq;
	char		vpid[VIDSZ+PIDSZ+1];
	char		rev[REVSZ+1];

	/*
	 * get inquiry data for vid/pid, and rev.
	 */
	(void) memset((void *) &InqReq, 0, sizeof(CHGR_INQ_PARAMS));
	(void) memset((void *) &Inq, 0, sizeof(CHGR_INQUIRY));
	InqReq.inqp_inq_ptr = (char *) &Inq;
	/* InqReq.inqp_inq_size = sizeof(CHGR_INQUIRY); */
	InqReq.inqp_inq_size = 36;
	InqReq.inqp_control = 0;
	(void) ioctl(fd, CHGRIOC_INQUIRY, (void *)&InqReq);

	strncpy(vpid, Inq.inq_vid, VIDSZ);
	strncpy(&vpid[VIDSZ], Inq.inq_pid, PIDSZ);
    	strntrm(vpid, PIDSZ+VIDSZ+1);
	strncpy(rev, Inq.inq_rev, REVSZ);
    	strntrm(rev, REVSZ+1);
 
#ifdef DEBUG
	fprintf(stdout, "%s Revision %s \n", vpid, rev);
	fprintf(stdout, "\tNumber of Drives     : %d\n",
	    (u_short) Addresses.adr_dt_number);
	fprintf(stdout, "\tFirst Drive          : %d\n",
	    (u_short) Addresses.adr_dt_address);
	fprintf(stdout, "\tNumber of Slots      : %d\n",
	    (u_short) Addresses.adr_st_number);
	fprintf(stdout, "\tFirst Slot           : %d\n",
	    (u_short) Addresses.adr_st_address);
	fprintf(stdout, "\tNumber of Transports : %d\n",
	    (u_short) Addresses.adr_mt_number);
	fprintf(stdout, "\tFirst Transport      : %d\n",
	    (u_short) Addresses.adr_mt_address);
	fprintf(stdout, "\tNumber of IE ports   : %d\n",
	    (u_short) Addresses.adr_ie_number);
	fprintf(stdout, "\tFirst IE port        : %d\n",
	    (u_short) Addresses.adr_ie_address);
	fprintf(stdout, "\tImport/Export        : %s\n",
	    (Addresses.adr_ie_number) ? "Supported" : "Not Supported");
	fprintf(stdout, "\tBarcode Labels       : %s\n",
	    (Barcode) ? "Supported" : "Not Supported");
#else
	fprintf(stdout, "%s Revision %s \n", vpid, rev);
	fprintf(stdout, "\tNumber of Drives     : %d\n",
	    (u_short) Addresses.adr_dt_number);
	fprintf(stdout, "\tNumber of Slots      : %d\n",
	    (u_short) Addresses.adr_st_number);
	fprintf(stdout, "\tNumber of Transports : %d\n",
	    (u_short) Addresses.adr_mt_number);
	fprintf(stdout, "\tNumber of IE ports   : %d\n",
	    (u_short) Addresses.adr_ie_number);
	fprintf(stdout, "\tImport/Export        : %s\n",
	    (Addresses.adr_ie_number) ? "Supported" : "Not Supported");
	fprintf(stdout, "\tBarcode Labels       : %s\n",
	    (Barcode) ? "Supported" : "Not Supported");
#endif /* DEBUG */

	return (0);
}

/*
 * Reset the target device
 */
int DoReset()
{
	/*
	 * This is a SCSI reset of the target device.  The
	 * device driver may have this ioctl restricted to
	 * to root users only.  If it is restricted to root
	 * then an EPERM error will be received by non-root
	 * users attempting this operation.
	 */
	if (ioctl(fd, CHGRIOC_RESETDEVICE, NULL) == -1) {
		perror(progname);
		if (errno == EPERM)
			(void) fprintf(stderr, "Device Reset can only be "
				"performed by root user\n");
		return(2);
	}
	return (0);
}

/*
 * Display inventory of changer elements and media
 */
int DoInventory()
{
	int			max_elems, nelems, bcount;
	CHGR_ESTATUS_PARAMS	StatusReq;
	CHGR_ESTATUS_HDR		*pStatusHdr;
	char			*s;

	max_elems = Addresses.adr_mt_number
			+ Addresses.adr_st_number
			+ Addresses.adr_ie_number
			+ Addresses.adr_dt_number;

	/*
	 * issue a Read Element Status to get status header only,
	 * in order to determine the correct allocation size needed.
	 */
	(void) memset((void *) &StatusReq, 0, sizeof(CHGR_ESTATUS_PARAMS));
	StatusReq.es_type = CHGR_EST_ALL;
	StatusReq.es_addr = 0;
	StatusReq.es_nelems = max_elems;
	StatusReq.es_tag = (Barcode && Labels)?1:0;
	StatusReq.es_status_size = STATUS_HDR_LENGTH;
	StatusReq.es_status_p = (char *) calloc(1, StatusReq.es_status_size);
	StatusReq.es_control = 0;
	if (ioctl(fd, CHGRIOC_ELEM_STATUS, (void *)&StatusReq) == -1) {
		perror(progname);
		GetSense();
		free(StatusReq.es_status_p);
		return (2);
	}
	pStatusHdr = (pCHGR_ESTATUS_HDR) StatusReq.es_status_p;

	s = (char *) &pStatusHdr->esh_nelems;
	S2TOI(s, nelems);
	S3TOI(pStatusHdr->esh_nbytes, bcount);
	bcount += sizeof(CHGR_ESTATUS_HDR);
#ifdef DEBUG
	(void) fprintf(stdout, "%d elements, %d bytes\n", nelems, bcount);
#endif
	free(StatusReq.es_status_p);
	StatusReq.es_type = CHGR_EST_ALL;
	StatusReq.es_addr = 0;
	StatusReq.es_nelems = nelems;
	StatusReq.es_tag = (Barcode && Labels)?1:0;
	StatusReq.es_status_size = bcount;
	StatusReq.es_status_p = (char *) calloc(1, bcount);
	if (ioctl(fd, CHGRIOC_ELEM_STATUS, (void *)&StatusReq) == -1) {
		perror(progname);
		GetSense();
		free(StatusReq.es_status_p);
		return(2);
	}

	/*
	 * Display element status
	 */
	PrintElementInfo(StatusReq.es_status_p);

	(void) fprintf(stdout, "\n");
	free(StatusReq.es_status_p);

	return (0);
}

/*
 * Search media for lables matching search string, and display status
 * for those media elements.
 */
int DoSearch()
{
	int			max_elems, nelems, bcount;
	char			*s;
	CHGR_SVTAG_PARAMS		*pSearchReq;
	CHGR_ESTATUS_PARAMS	StatusReq;
	CHGR_ESTATUS_HDR		*pStatusHdr;

	/*
	 * Check if device supports barcode labels
	 */
	if (!Barcode) {
		fprintf(stderr, "Device does not support labels\n");
		return(2);
	}
	Labels = 1;

	/*
	 * Check if a search string was given
	 */
	if (Search_Str == NULL) {
		(void) fprintf(stderr, "Search criteria not specified\n");
		(void) Usage(progname);
		return (2);
	}
#ifdef DEBUG
	(void) fprintf(stderr, "Search criteria = <%s>\n", Search_Str);
#endif

	max_elems = Addresses.adr_mt_number
			+ Addresses.adr_st_number
			+ Addresses.adr_ie_number
			+ Addresses.adr_dt_number;
#ifdef DEBUG
	(void) fprintf(stderr, "max_elems = %d (0x%x)\n", max_elems, max_elems);
#endif

	/*
	 * Seach media labels for name string
	 */
	(void) memset((void *)buf, 0, BUFSZ);
	pSearchReq = (pCHGR_SVTAG_PARAMS) buf;
	pSearchReq->svt_eaddr = 0;
	pSearchReq->svt_etype = CHGR_EST_ALL;
	pSearchReq->svt_acode = CHGR_ACODE_TRANS_SEARCH_PRI_NOSEQ;
	strcpy(pSearchReq->svt_info.ti_name, Search_Str);
	if (ioctl(fd, CHGRIOC_SEND_VOLTAG, (void *)buf) == -1) {
		perror(progname);
		GetSense();
		return (2);
	}

	/*
	 * issue a Request Volume Element Address to get status header only,
	 * in order to determine the correct allocation size needed.
	 */
	(void) memset((void *) &StatusReq, 0, sizeof(CHGR_ESTATUS_PARAMS));
	StatusReq.es_type = CHGR_EST_ALL;
	StatusReq.es_addr = 0;
	StatusReq.es_nelems = (u_short) max_elems;
	StatusReq.es_tag = 0;
	StatusReq.es_tag = (Barcode && Labels)?1:0;
	StatusReq.es_status_size = STATUS_HDR_LENGTH;
	StatusReq.es_status_p = (char *) calloc(1, StatusReq.es_status_size);
	if (ioctl(fd, CHGRIOC_REQ_VOL_ADDR, (void *)&StatusReq) == -1) {
		perror(progname);
		GetSense();
		free(StatusReq.es_status_p);
		return (2);
	}

	pStatusHdr= (pCHGR_ESTATUS_HDR) StatusReq.es_status_p;

	s = &(StatusReq.es_status_p[2]);
	S2TOI(s, nelems);
	S3TOI(pStatusHdr->esh_nbytes, bcount);
	bcount += STATUS_HDR_LENGTH;
	free(StatusReq.es_status_p);

#ifdef DEBUG
	(void) fprintf(stderr, "nelems = %d (0x%x)\n", nelems, nelems);
	(void) fprintf(stderr, "bcount = %d (0x%x) bytes\n", bcount, bcount);
#endif

	StatusReq.es_type = CHGR_EST_ALL;
	StatusReq.es_addr = 0;
	StatusReq.es_nelems = (u_short) nelems;
	StatusReq.es_tag = 0;
	StatusReq.es_tag = (Barcode && Labels)?1:0;
	StatusReq.es_status_size = bcount;
	StatusReq.es_status_p = (char *) calloc(1, StatusReq.es_status_size);
	if (ioctl(fd, CHGRIOC_REQ_VOL_ADDR, (void *)&StatusReq) == -1) {
		perror(progname);
		GetSense();
		free(StatusReq.es_status_p);
		return(2);
	}
	
	/*
	 * Display element status
	 */
	PrintElementInfo(StatusReq.es_status_p);

	(void) fprintf(stdout, "\n");
	free(StatusReq.es_status_p);

	return (0);
}

/*
 * Load a unit of media from a slot to a drive
 */
int DoSelect()
{
	CHGR_MOVE_PARAMS	*pLod;
	int		slot;

	(void) memset((void *) buf, 0, BUFSZ);
	pLod= (pCHGR_MOVE_PARAMS) buf;

	/*
	 * set which changer arm (MT element) to use
	 */
	pLod->mv_transport = 0;	/* use default MT */

	/*
	 * set the source address to move the medium from
	 * (must be a storage element).
	 */
	if (STVALID(STElement, Addresses)) {
		pLod->mv_source= STElement;
	} else {
		(void) fprintf(stderr, "Invalid source slot\n");
		(void) Usage(progname);
		return (2);
	}

	/*
	 * set the destination address to which DT element to load
	 * (must be a Data Transfer element).
	 */
	if (DTVALID(DTElement, Addresses)) {
		pLod->mv_dest = DTElement;
	} else {
		(void) fprintf(stderr, "Invalid destination drive\n");
		(void) Usage(progname);
		return (2);
	}

	/*
	 * Rotate the medium during transport
	 */
	if (Invert)
		pLod->mv_invert = 1;

	if (ioctl(fd, CHGRIOC_MOVE, (void *)buf) == -1) {
		perror(progname);
		GetSense();
		return(2);
	}
	return (0);
}


/*
 * Unload media from a drive and put into a storage slot.  If not
 * slot is specifed, try to put it back where it was loaded from.
 */
int DoDeselect()
{
	CHGR_MOVE_PARAMS		Uld;
	CHGR_ESTATUS_PARAMS	StatusReq;
	CHGR_ESTATUS_DESC		*pStatus;
	char			*s;
	int			prev;
	int			bcount;

	(void) memset((void *) &Uld, 0, sizeof(CHGR_MOVE_PARAMS));

	/*
	 * set which changer arm (MT element) to use
	 *
	Uld.mv_transport = 0;	* default MT element *
	 */

	/*
	 * set the source address to which DT element to unload
	 */
	if (DTVALID(DTElement, Addresses)) {
		Uld.mv_source= (u_short) DTElement;
	} else {
		(void) fprintf(stderr, "Invalid source drive\n");
		(void) Usage(progname);
		return (2);
	}

	/*
	 * set the destination storage address to unload medium to 
	 */
	if (STVALID(STElement, Addresses)) {
		Uld.mv_dest= (u_short) STElement;
	} else if (STElement == -1) {
		/*
		 * No destination storage address was specified, so
		 * we will try to determine the medium's previous
		 * location.
		 */
		(void)memset((void *) &StatusReq, 0, sizeof(CHGR_ESTATUS_PARAMS));
		(void) memset((void *)buf, 0, BUFSZ);
		StatusReq.es_type = CHGR_EST_DT;
		StatusReq.es_addr = DTElement;
		StatusReq.es_nelems = 1;
		StatusReq.es_tag = 0;
		StatusReq.es_status_size = STATUS_HDR_LENGTH;
		StatusReq.es_status_p = buf;
		if (ioctl(fd, CHGRIOC_ELEM_STATUS, (void *) &StatusReq) == -1) {
			perror(progname);
			GetSense();
			free(StatusReq.es_status_p);
			return (2);
		}

		s = &buf[5];
		S3TOI(s, bcount);
		StatusReq.es_status_size = bcount + STATUS_HDR_LENGTH;
		if (ioctl(fd, CHGRIOC_ELEM_STATUS, (void *) &StatusReq) == -1) {
			perror(progname);
			GetSense();
			free(StatusReq.es_status_p);
			return (2);
		}

		pStatus = (pCHGR_ESTATUS_DESC)
			&buf[STATUS_HDR_LENGTH + STATUS_PAGE_LENGTH];
		if (pStatus->esd_svalid) {
			s = (char *) &(pStatus->esd_src_addr);
			S2TOI(s, prev);
			Uld.mv_dest = (u_short) prev;
		} else {
			(void) fprintf(stderr, "No destination address "
				"specified and cannot determine media's "
				"previous location\n");
			(void) Usage(progname);
			return (2);
		}
	} else {
		(void) fprintf(stderr, 
		    "Invalid destination slot\n");
		(void) Usage(progname);
		return (2);
	}

	/*
	 * Rotate the medium during transport
	 */
	Uld.mv_invert = (Invert) ? 1 : 0;
	Uld.mv_control= 0;

	if (ioctl(fd, CHGRIOC_MOVE, (void *)&Uld) == -1) {
		perror(progname);
		GetSense();
		return(2);
	}
	return (0);
}

/*
 * Import media (from IE port) to a slot or drive
 */
int DoImport()
{
	CHGR_MOVE_PARAMS	*pImp;

	(void) memset((void *) buf, 0, BUFSZ);
	pImp= (pCHGR_MOVE_PARAMS) buf;

	/*
	 * Is there an import/export element for this device?
	 */
	if (Addresses.adr_ie_number == 0) {
		(void) fprintf(stderr,
		    "No import/export element for this device\n");
		return(2);
	}

	/*
	 * set which changer arm (MT element) to use
	pImp->mv_transport = 0;	* default MT element *
	 */

	/*
	 * set which import/export address (IE element) to use
	 */
	pImp->mv_source = (u_short) Addresses.adr_ie_address;

	/*
	 * set the destination address to import the medium to
	 * (ST or DT element).
	 */
	if (STVALID(STElement, Addresses)) {
		pImp->mv_dest = STElement;
	} else if (DTVALID(DTElement, Addresses)) {
		pImp->mv_dest = DTElement;
	} else {
		(void) fprintf(stderr, "Invalid destination element\n");
		(void) Usage(progname);
		return (2);
	}

	if (ioctl(fd, CHGRIOC_MOVE, (void *)buf) == -1) {
		perror(progname);
		GetSense();
		return(2);
	}
	return (0);
}

/*
 * Export medium from a slot or drive (to IE port).
 */
int DoExport()
{
	CHGR_MOVE_PARAMS	*pExp;

	(void) memset((void *) buf, 0, BUFSZ);
	pExp= (pCHGR_MOVE_PARAMS) buf;

	/*
	 * Is there an import/export element for this device?
	 */
	if (Addresses.adr_ie_number == 0) {
		(void) fprintf(stderr,
		    "No import/export element for this device\n");
		return(2);
	}

	/*
	 * set which changer arm (MT element) to use
	pExp->mv_transport = 0;	* default MT element *
	 */

	/*
	 * set the source address to export the medium from
	 * (can be either an ST or DT element).
	 */
	if (STVALID(STElement, Addresses)) {
		pExp->mv_source= STElement;
	} else if (DTVALID(DTElement, Addresses)) {
		pExp->mv_source= DTElement;
	} else {
		(void) fprintf(stderr, "Invalid source element\n");
		(void) Usage(progname);
		return (2);
	}

	/*
	 * set the destination address to the IE element for export
	 */
	pExp->mv_dest= (u_short) Addresses.adr_ie_address;

	if (ioctl(fd, CHGRIOC_MOVE, (void *)buf) == -1) {
		perror(progname);
		GetSense();
		return(2);
	}
	return (0);
}

/*
 * position medium transport element to a slot or drive.  If not
 * slot or drive position specified, park the transport arm to allow
 * easy access to media storage area.
 */
int DoPosition()
{
	CHGR_POS_PARAMS	*pPos;

	(void) memset((void *) buf, 0, BUFSZ);
	pPos= (pCHGR_POS_PARAMS) buf;

	/*
	 * set which changer arm (MT element) to use
	 */
	pPos->pos_transport= (u_short) Addresses.adr_mt_address;

	/*
	 * set the destination address to position in front of
	 */
	if ((STElement == -1) && (DTElement == -1)) {
		/*
		 * no destination address.  Just park it.
		 */
		pPos->pos_dest= pPos->pos_transport;
	} else if (STVALID(STElement, Addresses)) {
		pPos->pos_dest= STElement;
	} else if (DTVALID(DTElement, Addresses)) {
		pPos->pos_dest= DTElement;
	} else {
		(void) fprintf(stderr, "Invalid destination element\n");
		(void) Usage(progname);
		return (2);
	}

	/*
	 * Rotate the media handler
	 */
	if (Invert)
		pPos->pos_invert = 1;

	if (ioctl(fd, CHGRIOC_POSITION, (void *)buf) == -1) {
		perror(progname);
		GetSense();
		return(2);
	}
	return (0);
}

/*
 * move a unit of media
 */
int DoMove()
{
	CHGR_MOVE_PARAMS	*pMov;

	(void) memset((void *)buf, 0, BUFSZ);
	pMov = (pCHGR_MOVE_PARAMS) buf;

	/*
	 * set which changer arm (MT element) to use
	 */
	/*pMov->mv_transport = (u_short) Addresses.adr_mt_address;*/
	pMov->mv_transport = 0;

	/*
	 * set the source address to move the medium from
	 */
	if (STVALID(STElement, Addresses)) {
		pMov->mv_source = STElement;
	} else {
		(void) fprintf(stderr, "Invalid source slot\n");
		(void) Usage(progname);
		return (2);
	}

	/*
	 * set the destination address to move the medium to
	 */
	if (STVALID(STElementTwo, Addresses)) {
		pMov->mv_dest = STElementTwo;
	} else {
		(void) fprintf(stderr, "Invalid destination slot\n");
		(void) Usage(progname);
		return (2);
	}

	/*
	 * Rotate the medium during transport
	 */
	if (Invert)
		pMov->mv_invert = 1;

	if (ioctl(fd, CHGRIOC_MOVE, (void *)buf) == -1) {
		perror(progname);
		GetSense();
		return(2);
	}
	return (0);
}

/*
 * Perform a simple exchange of media
 */
int DoExchange()
{
	CHGR_EXCH_PARAMS	Exch;

	(void) memset((void *) &Exch, 0, sizeof(CHGR_EXCH_PARAMS));

	/*
	 * set the medium transport element to use
	 */
	Exch.exch_transport = 0;

	/*
	 * set the source address to move the medium from
	 */
	if (STVALID(STElement, Addresses)) {
		Exch.exch_source = STElement;
	} else {
		(void) fprintf(stderr, "Invalid source slot\n");
		(void) Usage(progname);
		return (2);
	}

	/*
	 * set the destination address to move the medium to
	 */
	if (STVALID(STElementTwo, Addresses)) {
		Exch.exch_first_dest = STElementTwo;
	} else {
		(void) fprintf(stderr, "Invalid destination slot\n");
		(void) Usage(progname);
		return (2);
	}

	/*
	 * set the second destination address 
	 */
	if (STElementThree == -1) {
		Exch.exch_second_dest = Exch.exch_source;
	} else if (STVALID(STElementThree, Addresses)) {
		Exch.exch_first_dest = STElementThree;
	} else {
		(void) fprintf(stderr, "Invalid destination slot\n");
		(void) Usage(progname);
		return (2);
	}

	/*
	 * Rotate media during exchange
	 */
	if (Invert) {
		Exch.exch_inv1 = Exch.exch_inv2 = 1;
	} else {
		Exch.exch_inv1 = Exch.exch_inv2 = 0;
	}

	/*
	 * set device specific control bits
	 */
	Exch.exch_control = 0;

	if (ioctl(fd, CHGRIOC_EXCHANGE, (void *)&Exch) == -1) {
		perror(progname);
		GetSense();
		return(2);
	}
	return (0);
}

/*
 * Lock the changer device to prevent media access/removal
 */
int DoLock()
{
	CHGR_LOCK_PARAMS	*pLock;

	(void) memset((void *)buf, 0, BUFSZ);
	pLock= (pCHGR_LOCK_PARAMS) buf;
	pLock->lock_lcode= CHGR_LC_LOCK;
	pLock->lock_control= 0;	/* lock door & IE port */
	if (ioctl(fd, CHGRIOC_LOCK, (void *)buf) == -1) {
		perror(progname);
		GetSense();
		return (2);
	}
	return (0);
}

/*
 * Unlock the changer device to allow media access/removal
 */
int DoUnlock()
{
	CHGR_LOCK_PARAMS	*pLock;

	(void) memset((void *)buf, 0, BUFSZ);
	pLock= (pCHGR_LOCK_PARAMS) buf;
	pLock->lock_lcode= CHGR_LC_UNLOCK;
	pLock->lock_control= 0;
	if (ioctl(fd, CHGRIOC_LOCK, (void *)buf) == -1) {
		perror(progname);
		GetSense();
		return (2);
	}
	return (0);
}

/*
 * retreive SCSI extended extended sense info for last scsi command
 * issued by the driver.
 */
void GetSense()
{
	CHGR_EXTENDED_SENSE	Sense;
	int			err;

	(void) memset((void *)&Sense, 0, sizeof(CHGR_EXTENDED_SENSE));
	if ((err = ioctl(fd, CHGRIOC_GETSENSE, &Sense)) == -1) {
		if (err != ENOMSG)
			perror(progname);
		return;
	}

#if (defined(_SOLARIS) && (!defined(_HPUX)))
	switch (Sense.es_key) {
	case KEY_NO_SENSE:
		fprintf(stdout, "Sense Key : 0x%x %s\n",
		    Sense.es_key, "No Sense");
		break;
	case KEY_NOT_READY:
		fprintf(stdout, "Sense Key : 0x%x %s\n",
		    Sense.es_key, "Not Ready");
		break;
	case KEY_HARDWARE_ERROR:
		fprintf(stdout, "Sense Key : 0x%x %s\n",
		    Sense.es_key, "Hardware Error");
		break;
	case KEY_ILLEGAL_REQUEST:
		fprintf(stdout, "Sense Key : 0x%x %s\n",
		    Sense.es_key, "Illegal Request");
		break;
	case KEY_UNIT_ATTENTION:
		fprintf(stdout, "Sense Key : 0x%x %s\n",
		    Sense.es_key, "Unit Attention");
		break;
	case KEY_ABORTED_COMMAND:
		fprintf(stdout, "Sense Key : 0x%x %s\n",
		    Sense.es_key, "Aborted Command");
		break;

	case KEY_RECOVERABLE_ERROR:
	case KEY_MEDIUM_ERROR:
	case KEY_DATA_PROTECT:
	case KEY_BLANK_CHECK:
	case KEY_VENDOR_UNIQUE:
	case KEY_COPY_ABORTED:
	case KEY_EQUAL:
	case KEY_VOLUME_OVERFLOW:
	default:
		(void) fprintf(stdout,"Sense Key : 0x%x %s\n",Sense.es_key," ");
	}
#else	/* _HPUX */
	(void) fprintf(stdout,"Sense Key : 0x%x %s\n",Sense.es_key," ");
#endif
	(void) fprintf(stdout, "Sense Data : "); 
	(void) fprintf(stdout, "Key= 0x%x, ASC= 0x%x, ASCQ= 0x%x,\n",
	    Sense.es_key, Sense.es_add_code, Sense.es_qual_code);
}

/*
 * Display element status info
 */
void PrintElementInfo(caddr_t status)
{
	CHGR_ESTATUS_HDR		*pStatusHdr;
	CHGR_ESTATUS_PAGE		*pStatusPage;
	CHGR_ESTATUS_DESC		*pStatusDesc;
	char			*s, *pCur;
	int			j, type, Vtag,
				bytes, count, statusbytes,
				descbytes;
	int			elem;

	/*
	 * print display header
	 */
	if (Barcode && Labels) {
		(void)fprintf(stdout,"Type      Addr  Status  Accessable    Label\n");
		(void)fprintf(stdout,"--------------------------------------------\n");
	} else {
		(void)fprintf(stdout,"Type      Addr  Status  Accessable \n");
		(void)fprintf(stdout,"-----------------------------------\n");
	}

	pCur = status;
	pStatusHdr = (pCHGR_ESTATUS_HDR) pCur;

	/*
	 * element status hdr info
	 */
#ifdef DEBUG
	(void) fprintf(stdout, "EStatus Hdr:\n");
	hex_print(status, STATUS_HDR_LENGTH);
#endif
	S3TOI(pStatusHdr->esh_nbytes, statusbytes);

	pCur +=  STATUS_HDR_LENGTH;	/* first status page */
	j= (statusbytes + STATUS_HDR_LENGTH);

	while (j > (pCur - status)) {
		pStatusPage = (pCHGR_ESTATUS_PAGE) pCur;

		/*
		 * element status page hdr info
		 */
#ifdef DEBUG
		(void) fprintf(stdout, "EStatus Page Hdr:\n");
		hex_print((char *)pStatusPage, STATUS_PAGE_LENGTH);
#endif
		type = (int)pStatusPage->esp_type;
		s = &pCur[2];
		S2TOI(s, descbytes);
		S3TOI(pStatusPage->esp_nbytes, bytes);
		if (pStatusPage->esp_pvoltag || pStatusPage->esp_avoltag) {
			Vtag = 1;
		} else {
			Vtag = 0;
		}
		count= bytes/descbytes;	/* number of elem descs for page */
		pCur += STATUS_PAGE_LENGTH;	/* 1st desc	 */

		while (count--) {
			pStatusDesc= (pCHGR_ESTATUS_DESC) pCur;

			/*
			 * element status descriptor info
			 */
#ifdef DEBUG
			(void) fprintf(stdout, "EStatus Desc:\n");
			hex_print((char *)pStatusDesc, STATUS_DECHGR_LENGTH);
#endif
			s = pCur;
			S2TOI(s, elem);

			switch(type) {
			case CHGR_EST_MT:
				fprintf(stdout, "Transport ");
				fprintf(stdout, "%3d   ",
				    elem - Addresses.adr_mt_address +1);
				break;

			case CHGR_EST_ST:
				fprintf(stdout, "Slot      ");
				fprintf(stdout, "%3d   ",
				    elem - Addresses.adr_st_address +1);
				break;

			case CHGR_EST_IE:
				fprintf(stdout, "Im/Export ");
				fprintf(stdout, "%3d   ",
				    elem - Addresses.adr_ie_address +1);
				break;

			case CHGR_EST_DT:
				fprintf(stdout, "Drive     ");
				fprintf(stdout, "%3d   ",
				    elem - Addresses.adr_dt_address +1);
				break;

			default:
				fprintf(stdout, "Unknown   ");
				fprintf(stdout, "%3d   ", elem);
			}

			fprintf(stdout, "%s    ",
			    (pStatusDesc->esd_full) ? "Full  ":"Empty ");
			fprintf(stdout, "%s   ",
			    (pStatusDesc->esd_access) ? "yes":"no ");
			if (Vtag && Barcode && Labels) {
				fprintf(stdout, "    %s",
						pStatusDesc->esd_pri_voltag);
			}
			printf("\n");

			pCur += descbytes;
		} /* end inner loop */
	} /* end outer loop */
}

/*
 * Trim trailing white space from a text string, and force string
 * to be null terminated.
 */
void strntrm(char *s, int len)
{
    register char *cp;

    cp = s + len - 1;
    *cp-- = 0;
    while (cp >= s && *cp == ' ')
    	*cp-- = 0;
}

int supports_barc()
{
	CHGR_INQ_PARAMS	InqReq;
	CHGR_INQUIRY	*Inq_p;
	int		rc = 0;

	/*
	 * get inquiry data
	 */
	(void) memset((void *) &InqReq, 0, sizeof(CHGR_INQ_PARAMS));
	(void) memset((void *)buf, 0, BUFSZ);
	InqReq.inqp_evpd= 0;
	InqReq.inqp_pcode= 0;		/* standard inquiry */
	InqReq.inqp_inq_ptr = (char *) &buf;
	InqReq.inqp_inq_size = sizeof(CHGR_INQUIRY);
	InqReq.inqp_control = 0;

	(void) ioctl(fd, CHGRIOC_INQUIRY, (void *)&InqReq);
	Inq_p = (CHGR_INQUIRY *) &buf;

	if (!strncmp("EXABYTE", Inq_p->inq_vid, 7)) {
		/*
		 * All Exabyte libraries will return 56 bytes of
		 * standard inq data.  For those that support barcode
		 * label readers, bit 0 of byte 56 will be set (except
		 * for the EXB-60 and EXB-120, which always support barcodes,
		 * but do not set this bit).
		 */
		if (((Inq_p->inq_vuniq[19] & 0x01) == 0x01) ||
		    (!strncmp("EXB-60", Inq_p->inq_pid, 6)) ||
		    (!strncmp("EXB-120", Inq_p->inq_pid, 7))) {
			rc = 1;
		}
	}
	return(rc);
} 

int GetAddrInfo(CHGR_ADDRS *addr)
{
	int	i;
	char	*s;

	(void) memset((void *) &buf, 0, sizeof(CHGR_ADDRS));
	if (ioctl(fd, CHGRIOC_ELEMENT_ADDRESSES, (void *)buf) == -1) {
		perror("Element Addresses");
		return (-1);
	}
	s = buf;
	S2TOI(s, addr->adr_mt_address);
	s = &buf[2];
	S2TOI(s, addr->adr_mt_number);
	s = &buf[4];
	S2TOI(s, addr->adr_st_address);
	s = &buf[6];
	S2TOI(s, addr->adr_st_number);
	s = &buf[8];
	S2TOI(s, addr->adr_ie_address);
	s = &buf[10];
	S2TOI(s, addr->adr_ie_number);
	s = &buf[12];
	S2TOI(s, addr->adr_dt_address);
	s = &buf[14];
	S2TOI(s, addr->adr_dt_number);
	
	return(0);
}

#ifdef DEBUG
#define CDB_GROUP0	6

int DoRezero()
{
	CHGR_USCSI_CMD	scmd;
	char		cmdblk[CDB_GROUP0];

	(void) memset((void *)&scmd, 0, sizeof(scmd));
	(void) memset((void *)cmdblk, 0, CDB_GROUP0);

	cmdblk[0] = 0x01;
	scmd.uscsi_flags = 0;
	scmd.uscsi_timeout = 30;
	scmd.uscsi_cdb = cmdblk;
	scmd.uscsi_bufaddr = NULL;
	scmd.uscsi_buflen = 0;
	scmd.uscsi_cdblen = CDB_GROUP0;

	if (ioctl(fd, USCSICMD, (void *)&scmd) == -1) {
		perror(progname);
		GetSense();
		return (2);
	}
	return(0);
}

int DoInquiry()
{
	CHGR_INQ_PARAMS	InqReq;
	CHGR_INQUIRY	*Inq_p;
	char		vpid[VIDSZ+PIDSZ+1];
	char		rev[REVSZ+1];
	int		err = 0;

	/*
	 * get inquiry data
	 */
	(void) memset((void *) &InqReq, 0, sizeof(CHGR_INQ_PARAMS));
	(void) memset((void *)buf, 0, BUFSZ);
	InqReq.inqp_evpd= 0;
	InqReq.inqp_pcode= 0;		/* standard inquiry */
	InqReq.inqp_inq_ptr = (char *) &buf;
	InqReq.inqp_inq_size = sizeof(CHGR_INQUIRY);
	InqReq.inqp_control = 0;
	if (ioctl(fd, CHGRIOC_INQUIRY, (void *)&InqReq) == -1) {
		perror(progname);
		GetSense();
		err = 2;
	}
	Inq_p = (CHGR_INQUIRY *) &buf;
	strncpy(vpid, Inq_p->inq_vid, VIDSZ);
	strncpy(&vpid[VIDSZ], Inq_p->inq_pid, PIDSZ);
    	strntrm(vpid, PIDSZ+VIDSZ+1);
	strncpy(rev, Inq_p->inq_rev, REVSZ);
    	strntrm(rev, REVSZ+1);

	printf("Peripheral device type            = %d\n", Inq_p->inq_dtype);
	printf("Peripheral qualifer               = %d\n", Inq_p->inq_pqual);
	printf("Device type qualifer              = %d\n", Inq_p->inq_dtq);
	printf("Removable media                   = %d\n", Inq_p->inq_rmb);
	printf("ANSI version                      = %d\n", Inq_p->inq_ansi);
	printf("ECMA version                      = %d\n", Inq_p->inq_ecma);
	printf("ISO version                       = %d\n", Inq_p->inq_iso);
	printf("Response data format              = %d\n", Inq_p->inq_rdf);
	printf("Supports Term IO proc msg         = %d\n", Inq_p->inq_trmiop);
	printf("Supports Async event notification = %d\n", Inq_p->inq_aenc);
	printf("Additional Inquiry length         = %d\n", Inq_p->inq_addlen);
	printf("Supports Soft reset option        = %d\n", Inq_p->inq_sftre);
	printf("Supports cmd queueing             = %d\n", Inq_p->inq_cmdque);
	printf("Supports linked cmds              = %d\n", Inq_p->inq_linked);
	printf("Supports sync data xfers          = %d\n", Inq_p->inq_sync);
	printf("Supports 16 bit wide bus xfers    = %d\n", Inq_p->inq_wbus16);
	printf("Supports 32 bit wide bus xfers    = %d\n", Inq_p->inq_wbus32);
	printf("Supports relative addressing      = %d\n", Inq_p->inq_reladdr);
	printf("Vendor/Product ID string          = %s\n", vpid);
	printf("Revision string                   = %s\n", rev);

	if (Inq_p->inq_addlen > 36) {
		printf("Vendor Specific :\n");
		hex_print((char *)Inq_p->inq_vuniq, 20);
	}

	return(err);
} 

/*
 * Print a buffer readably
 */
void hex_print(char *cp, int len)
{
	int i, j=0;

	(void) fprintf(stdout, "[%d]\t", (int) cp);
	for (i = 0; i < len; i++) {
		(void) fprintf(stdout, "%x ", cp[i]);
		if ((j++ == 7) && (i+1 < len)) {
			(void) fprintf(stdout, "\n[%d]\t", (int) &cp[i]);
			j = 0;
		}
	}
	(void) fprintf(stdout, "\n");
}

#endif /*DEBUG*/
