#pragma	ident	"@(#)display.c 1.0	11/10/95 Exabyte Corp"

/*
 * display.c - set the display message on Exabyte CHS library unit 
 *
 * This file is a product of Exabyte Corporation and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify this file without charge, but are not authorized to
 * license or distribute it to anyone else except as part of a product
 * or program developed by the user.
 *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 * This file is provided with no support and without any obligation on the
 * part of Exabyte Corp. to assist in its use, correction, modification or
 * enhancement.
 *
 * EXABYTE CORP. SHALL HAVE NO LIABILITY WITH RESPECT TO THE INFRINGEMENT
 * OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS FILE OR ANY PART
 * THEREOF.
 *
 * In no event will Exabyte Corporation be liable for any lost revenue or
 * profits or other special, indirect and consequential damages, even if
 * Exabyte has been advised of the possibility of such damages.
 *
 * Exabyte Corporation
 * 1685 38th Street
 * Boulder, Colorado  80301
 */

#if defined(solaris) || defined(sparc) || defined(i386) 
#include <sys/isa_defs.h>
#endif

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/scsi/scsi.h>
#include <errno.h>
#include <sys/chgrio.h>


typedef u_long	flag_t;

/*
 * Local Structures and defines
 */
#define VERS_MAJOR	0
#define VERS_MINOR	0
#define	BUFSZ		512
#define MAX_MSG_LEN	80
#define LED_MSG_LEN	60
#define LCD_MSG_LEN	80	/* 4 lines, 20 chars per line */

typedef struct exb_vu_page {
	CHGR_MODE_PARAM_HDR	param_hdr;
	CHGR_MODE_PAGE_HDR	page_hdr;
#if defined(_BIT_FIELDS_LTOH)
	u_char	exbvu_mdc	:2,
		exbvu_nrdc	:1,
		exbvu_resv	:1,
		exbvu_nbl	:1,
		exbvu_parity	:1,
		exbvu_unit	:1,
		exbvu_ainit	:1;
#elif defined(_BIT_FIELDS_HTOL)
	u_char	exbvu_ainit	:1,
		exbvu_unit	:1,
		exbvu_parity	:1,
		exbvu_nbl	:1,
		exbvu_resv	:1,
		exbvu_nrdc	:1,
		exbvu_mdc	:2;
#else
#error One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif /( _BIT_FIELDS_LTOH */
	u_char	exbvu_max_par_ret;
	char	exbvu_disp_msg[60];
} EXBVU_PAGE;
#define EXBVU_PAGE_LEN	sizeof(struct exb_vu_page)

typedef struct lcd_page {
	CHGR_MODE_PARAM_HDR	param_hdr;
	CHGR_MODE_PAGE_HDR	page_hdr;
#if defined(_BIT_FIELDS_LTOH)
	u_char	lcd_write_l4	:1,
		lcd_write_l3	:1,
		lcd_write_l2	:1,
		lcd_write_l1	:1,
		lcd_rsvd	:4;
#elif defined(_BIT_FIELDS_HTOL)
	u_char	lcd_rsvd	:4,
		lcd_write_l1	:1,
		lcd_write_l2	:1,
		lcd_write_l3	:1,
		lcd_write_l4	:1;
#else
#error One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif /( _BIT_FIELDS_LTOH */
	u_char	lcd_reserved;
	char	lcd_disp_l1[20];
	char	lcd_disp_l2[20];
	char	lcd_disp_l3[20];
	char	lcd_disp_l4[20];
} LCD_PAGE;
#define LCD_PAGE_LEN	sizeof(struct lcd_page)


/*
 * Known media changer devices
 */
#define UNKNOWN		0
#define EXB60		0x01
#define	EXB120		0x02
#define EXB10e		0x03
#define EXB210		0x04
#define EXB440		0x05

/*
 * Message display codes
 */
#define FLASHING	0x00
#define STEADY		0x01
#define SCROLLING	0x02

/*
 * Local Macros
 */
#define STVALID(x, y)	(int)(((int)x >= (int)y.adr_st_address) && \
			((int)x < (int)(y.adr_st_address + y.adr_st_number)))
#define DTVALID(x,y)	(int)(((int)x >= (int)y.adr_dt_address) && \
			((int)x < (int)(y.adr_dt_address + y.adr_dt_number)))
#define MTVALID(x,y)	(int)(((int)x >= (int)y.adr_mt_address) && \
			((int)x < (int)(y.adr_mt_address + y.adr_mt_number)))
#define IEVALID(x,y)	(int)(((int)x >= (int)y.adr_ie_address) && \
			((int)x < (int)(y.adr_ie_address + y.adr_ie_number)))
#define ITOS2(i, s)	s[0]= (u_char)(((int)i & 0xff00) >> 8); \
			s[1]= (u_char)((int)i & 0x00ff)
#define S2TOI(s, i)	i = (int) (((int)s[0]) << 8) + (int)s[1]
#define ITOS3(i, s)	s[0]= (u_char)(((int)i & 0xff0000) >> 16); \
			s[1]= (u_char)(((int)i & 0x00ff00) >> 8); \
			s[2]= (u_char)((int)i & 0x0000ff)
#define S3TOI(s, i)	i = (int) (((int)s[0]) << 16) + \
				  (((int)s[1]) <<  8) + \
				    (int)s[2]

/*
 * Global Data
 */
char	*progname;		/* the name of this program */
char	dev[] = "/dev/chgr0";	/* default device name */
char	*devname = dev;		/* medium changer device */
char	buf[BUFSZ];		/* command data buffer */
flag_t	message_flg = 0;
flag_t	restore_flg = 0;
flag_t	steady_flg = 0;

/*
 * Function Declarations
 */
void usage(char *);
int get_vpid(int, char *);
int get_message(int, int, char *);
int set_message(int, int, char *);
int mode_sense(int, int, int, void *, int);
int mode_select(int, void *, int);
void strntrm(char *, int);
#ifdef DEBUG
void hex_print(char *, int);
#endif /*DEBUG*/


int main(int argc, char **argv)
{
	extern char	*optarg;
	extern int	optind;
	char		msg[MAX_MSG_LEN-1],
			id_str[VIDSZ+PIDSZ+1];
	flag_t		flags = 0;
	int		id, fd, c, err;


	progname = argv[0];	/* name of this program */
	memset(msg, 0, sizeof(msg));

	/*
	 * Check for command line arguments and flags
	 */
	while ((c=getopt(argc, argv, "?f:hm:rs")) != EOF) {
		switch(c) {
			case 'f':
			/*
			 * A device name has been specified on commandline,
			 * use it instead of the default.
			 */
			devname= optarg;
			break;

			case 'm':
			/*
			 * a new 'Ready' display message has been specified
			 * on the command line.  Set the device's 'Ready'
			 * message to this new string.
			 */
			message_flg = 1;
			strncpy(msg, optarg, MAX_MSG_LEN);
			msg[MAX_MSG_LEN] = 0;
			break;

			case 'r':
			/*
			 * restore default message.
			 */
			restore_flg = 0x01;
			break;

			case 's':
			/*
			 * keep message steady (i.e., not flashing or
			 * scrolling.
			 */
			steady_flg = 1;
			break;

			case 'h':
			case '?':
			default:
				(void) usage(progname);
				return(-1);

		}
	}

	/*
	 * Open the device.
	 */
	if ((fd = open(devname, O_RDWR)) == -1) {
		perror(progname);
		exit(1);
	}

	/*
	 * Process user request(s)
	 */
	if (get_vpid(fd, id_str) != 0) {
		(void) fprintf(stderr, "%s: Could not identify device %s\n",
			progname, devname);
		return(-1);
	}
	(void) fprintf(stdout, "Device %s is a %s\n", devname, id_str);

	if (!strcmp("EXABYTE EXB-60", id_str)) {
		id = EXB60;
	} else if (!strcmp("EXABYTE EXB-120", id_str)) {
		id = EXB120;
	} else if (!strcmp("EXABYTE EXB-210", id_str)) {
		id = EXB210;
	}

	if (message_flg) {
		/*
		 * New message string specified on cmd line, set
		 * device ready message.
		 */
		(void) fprintf(stdout,
			"Setting device ready message to <%s>...\n", msg);
		if (set_message(fd, id, msg) != 0) {
			fprintf(stderr, "Error setting new message\n");
		} else {
			(void) fprintf(stdout, "...Done\n");
		}
	} else {
		/*
		 * No message string specified on cmd line, just get
		 * current message string from device and display it.
		 */
		if (get_message(fd, id, msg) != 0) {
			fprintf(stderr, "Error retreiving message\n");
		} else {
			if (restore_flg) {
				if (set_message(fd, id, msg) != 0) {
					(void) fprintf(stderr,
				    	"Error restoring default message <%s>\n",
				    	msg);
				}
			}
			(void) fprintf(stdout,
			    "The current message string is <%s>\n", msg);
		}
	}
	

	/*
	 * Close the device and exit
	 */
	(void) close(fd);
	return(err);
}

/*
 * Display program usage info.
 */
void usage(char *name)
{
	(void) fprintf(stderr, "%s Version %d.%d\n", name, VERS_MAJOR,
		VERS_MINOR);
	(void) fprintf(stderr,
		"\tUsage: %s [-f device] [-m msg_string] [-r] [-h] [-?]\n", name);
}

/*
 * Get the medium changer device's Vendor/Product ID string from the
 * device Inquiry data.
 */
int get_vpid(int fd, char *vpid)
{
	CHGR_INQ_PARAMS	InqReq;
	CHGR_INQUIRY	*Inq_p;
	int		err = 0;
#ifdef DEBUG
	char		rev[REVSZ+1];
#endif /*DEBUG*/

	/*
	 * get inquiry data
	 */
	InqReq.inqp_evpd= 0;
	InqReq.inqp_pcode= 0;		/* standard inquiry */
	InqReq.inqp_inq_ptr = (char *) &buf;
	InqReq.inqp_inq_size = sizeof(CHGR_INQUIRY);
	InqReq.inqp_control = 0;

	if ((err= ioctl(fd, CHGRIOC_INQUIRY, (void *)&InqReq)) != 0) {
		perror(progname);
		return(err);
	}

	Inq_p = (CHGR_INQUIRY *) &buf;
	strncpy(vpid, Inq_p->inq_vid, VIDSZ);
	strncpy(&vpid[VIDSZ], Inq_p->inq_pid, PIDSZ);
    	strntrm(vpid, PIDSZ+VIDSZ+1);

#ifdef DEBUG
	strncpy(rev, Inq_p->inq_rev, REVSZ);
    	strntrm(rev, REVSZ+1);

	printf("Peripheral device type            = %d\n", Inq_p->inq_dtype);
	printf("Peripheral qualifer               = %d\n", Inq_p->inq_pqual);
	printf("Device type qualifer              = %d\n", Inq_p->inq_dtq);
	printf("Removable media                   = %d\n", Inq_p->inq_rmb);
	printf("ANSI version                      = %d\n", Inq_p->inq_ansi);
	printf("ECMA version                      = %d\n", Inq_p->inq_ecma);
	printf("ISO version                       = %d\n", Inq_p->inq_iso);
	printf("Response data format              = %d\n", Inq_p->inq_rdf);
	printf("Supports Term IO proc msg         = %d\n", Inq_p->inq_trmiop);
	printf("Supports Async event notification = %d\n", Inq_p->inq_aenc);
	printf("Additional Inquiry length         = %d\n", Inq_p->inq_addlen);
	printf("Supports Soft reset option        = %d\n", Inq_p->inq_sftre);
	printf("Supports cmd queueing             = %d\n", Inq_p->inq_cmdque);
	printf("Supports linked cmds              = %d\n", Inq_p->inq_linked);
	printf("Supports sync data xfers          = %d\n", Inq_p->inq_sync);
	printf("Supports 16 bit wide bus xfers    = %d\n", Inq_p->inq_wbus16);
	printf("Supports relative addressing      = %d\n", Inq_p->inq_reladdr);
	printf("Vendor/Product ID string          = %s\n", vpid);
	printf("Revision string                   = %s\n", rev);

	if (Inq_p->inq_addlen > 31) {
		printf("Vendor Specific :\n");
		hex_print((char *)Inq_p->inq_vuniq, sizeof(Inq_p->inq_vuniq));
	}
#endif /*DEBUG*/

	return(err);
} 

/*
 * Get display message
 */
int get_message(int fd, int dev_id, char *mesg_str)
{
	char		buf[BUFSZ];
	EXBVU_PAGE	*vu_page;
	LCD_PAGE	*lcd_page;
	void		*page;
	int		pg_len,
			pg_ctrl,
			pg_code,
			err = 0;
		

	switch(dev_id) {
		case EXB60:
		case EXB120:
			/*
			 * Use vendor unique mode page (00h)
			 */
			vu_page = (EXBVU_PAGE *) buf;
			page = (void *) vu_page;
			pg_len = EXBVU_PAGE_LEN;
			pg_code = 0x00;

			if (restore_flg) pg_ctrl = 0x02;
			else pg_ctrl = 0;

			err= mode_sense(fd, pg_ctrl, pg_code, page, pg_len);
			if (err != 0) {
				(void) fprintf(stderr, "mode_sense failed\n");
				perror(progname);
			} else {
				strcpy(mesg_str, vu_page->exbvu_disp_msg);
			}
			break;

		case EXB10e:
		case EXB210:
		case EXB440:
			/*
			 * Use LCD mode page (22h)
			 */
			lcd_page = (LCD_PAGE *) buf;
			page = (void *) lcd_page;
			pg_len = LCD_PAGE_LEN;
			pg_code = 0x22;

			if (restore_flg) pg_ctrl = 0x02;
			else pg_ctrl = 0;

			err= mode_sense(fd, pg_ctrl, pg_code, page, pg_len);
			if (err != 0) {
				(void) fprintf(stderr, "mode_sense failed\n");
				perror(progname);
			} else {
				strncpy(mesg_str, lcd_page->lcd_disp_l1, LCD_MSG_LEN);
/***
				strncpy(&mesg_str[20],lcd_page->lcd_disp_l2,20);
				strncpy(&mesg_str[20],lcd_page->lcd_disp_l3,20);
				strncpy(&mesg_str[20],lcd_page->lcd_disp_l4,20);
***/
			}
			break;

		case UNKNOWN:
		default:
			(void) fprintf(stderr, "Unrecognized device\n");
			return(-1);
	}
	return(err);
}

/*
 * Set display message
 */
int set_message(int fd, int dev_id, char *mesg_str)
{
	char		buf[BUFSZ];
	EXBVU_PAGE	*vu_page;
	LCD_PAGE	*lcd_page;
	void		*page;
	int		len,
			pg_len,
			pg_ctrl,
			pg_code,
			err = 0;


	switch(dev_id) {
		case EXB60:
		case EXB120:
			/*
			 * Use vendor unique mode page (00h)
			 */
			vu_page = (EXBVU_PAGE *) buf;
			page = (void *) vu_page;
			pg_len = EXBVU_PAGE_LEN;
			pg_code = 0x00;
			memset(page, 0, pg_len);

			vu_page->page_hdr.mph_pcode = (u_char) pg_code;
			vu_page->page_hdr.mph_page_length = (u_char)
			    (pg_len - MODE_HDR_SIZE - PAGE_HDR_SIZE); 
			strncpy(vu_page->exbvu_disp_msg, mesg_str, LCD_MSG_LEN);

			if ((int)strlen(mesg_str) > 8) {
				vu_page->exbvu_mdc = 0x02;
			} else {
				if (steady_flg) vu_page->exbvu_mdc = 0x01;
			}
			break;

		case EXB10e:
		case EXB210:
		case EXB440:
			/*
			 * Use LCD mode page (22h)
			 */
			lcd_page = (LCD_PAGE *) buf;
			page = (void *) lcd_page;
			pg_len = LCD_PAGE_LEN;
			pg_code = 0x22;
			memset(page, 0, pg_len);

			lcd_page->page_hdr.mph_pcode = (u_char) pg_code;
			lcd_page->page_hdr.mph_page_length = (u_char)
			    (pg_len - MODE_HDR_SIZE - PAGE_HDR_SIZE); 

			if (!restore_flg) {
				len = strlen(mesg_str);
				if (len > 0)
					lcd_page->lcd_write_l1 = 1;
				if (len > 20)
					lcd_page->lcd_write_l2 = 1;
				if (len > 40)
					lcd_page->lcd_write_l3 = 1;
				if (len > 60)
					lcd_page->lcd_write_l4 = 1;
				strncpy(lcd_page->lcd_disp_l1, mesg_str,
				    LCD_MSG_LEN);
			} else {
				strncpy(lcd_page->lcd_disp_l1, mesg_str, 32);
			}
			
			break;
			
		case UNKNOWN:
		default:
			(void) fprintf(stderr, "Unrecognized device\n");
			return(-1);
	}

	if (err= mode_select(fd, page, pg_len) != 0) {
		(void) fprintf(stderr, "mode_select failed\n");
		perror(progname);
	}
	return(err);
}

/*
 * Mode sense
 */
int mode_sense(int fd, int pc, int page, void *data_buf, int data_len)
{
	CHGR_USCSI_CMD	scmd;
	char	cmdblk[CDB_GROUP0];
	int		err;

	(void) memset((void *) &scmd, 0, sizeof(scmd));
	(void) memset((void *) cmdblk, 0, CDB_GROUP0);

	cmdblk[0] = CHGR_SCMD_MODESENSE;
	cmdblk[1] |= 0x08;	/* Disable Block Descriptors */
	cmdblk[2] = (unsigned char) (((pc & 0x03) << 6) | (page & 0x3F));
	cmdblk[4] = (unsigned char) data_len;

	scmd.uscsi_flags = USCSI_READ;
	scmd.uscsi_timeout = 30;
	scmd.uscsi_cdb = cmdblk;
	scmd.uscsi_bufaddr = data_buf;
	scmd.uscsi_buflen = (unsigned int) data_len;
	scmd.uscsi_cdblen = CDB_GROUP0;

	err = ioctl(fd, USCSICMD, &scmd);
	return(err);
}

/*
 * Mode select
 */
int mode_select(int fd, void *data_buf, int data_len)
{
	CHGR_USCSI_CMD	scmd;
	char	cmdblk[CDB_GROUP0];
	CHGR_EXTENDED_SENSE	*sense;
	int		err;

	(void) memset((void *) &scmd, 0, sizeof(scmd));
	(void) memset((void *) cmdblk, 0, CDB_GROUP0);

	cmdblk[0] = CHGR_SCMD_MODESELECT;
	cmdblk[1] |= 0x11;	/* save page */
	cmdblk[4] = (unsigned char) data_len;

	scmd.uscsi_flags = USCSI_WRITE;
	scmd.uscsi_timeout = 30;
	scmd.uscsi_cdb = cmdblk;
	scmd.uscsi_bufaddr = data_buf;
	scmd.uscsi_buflen = (unsigned int) data_len;
	scmd.uscsi_cdblen = CDB_GROUP0;

	err = ioctl(fd, USCSICMD, &scmd);
	return(err);
}

/*
 * Trim trailing white space from a text string, and force string
 * to be null terminated.
 */
void strntrm(char *s, int len)
{
    register char *cp;

    cp = s + len - 1;
    *cp-- = 0;
    while (cp >= s && *cp == ' ')
    	*cp-- = 0;
}

#ifdef DEBUG
/*
 * Print a buffer readably
 */
void hex_print(char *cp, int len)
{
	int i, j=0;

	(void) fprintf(stdout, "[%d]\t", (int) cp);
	for (i = 0; i < len; i++) {
		(void) fprintf(stdout, "%x ", cp[i]);
		if ((j++ == 7) && (i+1 < len)) {
			(void) fprintf(stdout, "\n[%d]\t", (int) &cp[i]);
			j = 0;
		}
	}
	(void) fprintf(stdout, "\n");
}
#endif /*DEBUG*/

